"""Config file for comfyui-speedup, which will automatically"""
import os
import sys
from pathlib import Path

from onediff.infer_compiler.backends.oneflow.utils.version_util import (
    is_community_version,
)

# Delete the diffusers module
if "diffusers" in sys.modules:
    del sys.modules["diffusers"]

# Set up paths
ONEDIFF_QUANTIZED_OPTIMIZED_MODELS = "onediff_quant"
COMFYUI_ROOT = os.getenv("COMFYUI_ROOT")

custom_nodes_path = os.path.join(COMFYUI_ROOT, "custom_nodes")
infer_compiler_registry_path = os.path.join(
    os.path.dirname(__file__), "infer_compiler_registry"
)

# Add paths to sys.path if not already there
if custom_nodes_path not in sys.path:
    sys.path.append(custom_nodes_path)

if infer_compiler_registry_path not in sys.path:
    sys.path.append(infer_compiler_registry_path)

# infer_compiler_registry/register_comfy
import register_comfy  # load plugins

_USE_UNET_INT8 = not is_community_version()
if _USE_UNET_INT8:
    # infer_compiler_registry/register_onediff_quant
    import register_onediff_quant  # load plugins
    from folder_paths import folder_names_and_paths, models_dir, supported_pt_extensions

    unet_int8_model_dir = Path(models_dir) / "unet_int8"
    unet_int8_model_dir.mkdir(parents=True, exist_ok=True)
    folder_names_and_paths["unet_int8"] = (
        [str(unet_int8_model_dir)],
        supported_pt_extensions,
    )

    opt_models_dir = Path(models_dir) / ONEDIFF_QUANTIZED_OPTIMIZED_MODELS
    opt_models_dir.mkdir(parents=True, exist_ok=True)

    folder_names_and_paths[ONEDIFF_QUANTIZED_OPTIMIZED_MODELS] = (
        [str(opt_models_dir)],
        supported_pt_extensions,
    )



# ComfyUI/comfy/ldm/modules/diffusionmodules/model.py

import oneflow as torch
from onediff.infer_compiler.backends.oneflow.transform import transform_mgr

onediff_comfy = transform_mgr.transform_package("comfy")

AttnBlock_porxy = onediff_comfy.ldm.modules.diffusionmodules.model.AttnBlock


# https://github.com/comfyanonymous/ComfyUI/blob/b4e915e74560bd2c090f9b4ed6b73b0781b7050e/comfy/ldm/modules/diffusionmodules/model.py#L243
class AttnBlock(AttnBlock_porxy):
    def forward(self, x):
        h_ = x
        h_ = self.norm(h_)
        q = self.q(h_)
        k = self.k(h_)
        v = self.v(h_)
        B, C, _, _ = x.shape
        # compute attention
        q, k, v = map(
            lambda t: t.reshape(B, 1, C, -1).transpose(2, 3).contiguous(),
            (q, k, v),
        )

        _, _, _, head_dim = q.shape
        is_causal = False
        out = torch._C.fused_multi_head_attention_inference_v2(
            query=q,
            query_layout="BHMK",
            query_head_size=head_dim,
            key=k,
            key_layout="BHMK",
            value=v,
            value_layout="BHMK",
            output_layout="BM(HK)",
            causal=is_causal,
        )

        out = out.transpose(1, 2).reshape(B, C, -1)
        out = out.reshape_as(x)

        out = self.proj_out(out)

        return x + out



import itertools
from typing import Callable, Dict, List, NamedTuple, Tuple, Union


def create_generator_registry():
    # Dictionary to hold registered constructors
    generator_registry: Dict[str, Callable] = {}

    def register(workflow_path: Union[List[str], str]) -> Callable:
        def decorator(generator_function: Callable) -> Callable:
            if isinstance(workflow_path, (List, Tuple)):
                for workflow in workflow_path:
                    generator_registry[workflow] = generator_function
            else:
                generator_registry[workflow_path] = generator_function
            return generator_function

        return decorator

    def dispatch(workflow_path: Union[List[str], str], *args, **kwargs) -> NamedTuple:
        if isinstance(workflow_path, (List, Tuple)):
            return itertools.chain(
                *[dispatch(w, *args, **kwargs) for w in workflow_path]
            )
        else:
            generator = generator_registry.get(workflow_path)
            if generator is None:
                raise ValueError(f"No generator registered for {workflow_path}")
            return generator(workflow_path, *args, **kwargs)

    return register, dispatch



import importlib.metadata
from typing import Optional

import oneflow as torch
import oneflow.nn as nn
import oneflow.nn.functional as F

from onediff.infer_compiler.backends.oneflow.transform import transform_mgr
from packaging import version

diffusers_of = transform_mgr.transform_package("diffusers")
StableCascadeUnet_OF_CLS = (
    diffusers_of.pipelines.stable_cascade.modeling_stable_cascade_common.StableCascadeUnet
)

ResBlockStageB = (
    diffusers_of.pipelines.wuerstchen.modeling_wuerstchen_diffnext.ResBlockStageB
)
AttnBlock = diffusers_of.pipelines.wuerstchen.modeling_wuerstchen_common.AttnBlock
TimestepBlock = (
    diffusers_of.pipelines.wuerstchen.modeling_wuerstchen_common.TimestepBlock
)

num_overflow_up_blocks = 1


class StableCascadeUnet_OF(StableCascadeUnet_OF_CLS):
    def of_up_decode(self, level_outputs, r_embed, clip):
        x = level_outputs[0]
        block_group = zip(self.up_blocks, self.up_upscalers, self.up_repeat_mappers)
        for i, (up_block, upscaler, repmap) in enumerate(block_group):
            if i == len(self.up_blocks) - num_overflow_up_blocks:
                x = x.to(torch.bfloat16)
                r_embed = r_embed.to(torch.bfloat16)
                clip = clip.to(torch.bfloat16)
            for j in range(len(repmap) + 1):
                for k, block in enumerate(up_block):
                    block_class = block.__class__
                    if isinstance(block, ResBlockStageB):
                        skip = level_outputs[i] if k == 0 and i > 0 else None
                        if skip is not None and (
                            x.size(-1) != skip.size(-1) or x.size(-2) != skip.size(-2)
                        ):
                            x = torch.nn.functional.interpolate(
                                x.float(),
                                skip.shape[-2:],
                                mode="bilinear",
                                align_corners=True,
                            )
                        if (
                            skip is not None
                            and i >= len(level_outputs) - num_overflow_up_blocks
                        ):
                            skip = skip.to(torch.bfloat16)
                        x = block(x, skip)
                    elif isinstance(block, AttnBlock):
                        x = block(x, clip)
                    elif isinstance(block, TimestepBlock):
                        x = block(x, r_embed)
                    else:
                        x = block(x)
                if j < len(repmap):
                    x = repmap[j](x)
            x = upscaler(x)
        return x

    def forward(
        self,
        x,
        r,
        clip_text_pooled,
        clip_text=None,
        clip_img=None,
        effnet=None,
        pixels=None,
        sca=None,
        crp=None,
    ):
        if pixels is None:
            pixels = x.new_zeros(x.size(0), 3, 8, 8)

        # Process the conditioning embeddings
        r_embed = self.gen_r_embedding(r)
        for c in self.config.t_conds:
            if c == "sca":
                cond = sca
            elif c == "crp":
                cond = crp
            else:
                cond = None
            t_cond = cond or torch.zeros_like(r)
            r_embed = torch.cat([r_embed, self.gen_r_embedding(t_cond)], dim=1)
        clip = self.gen_c_embeddings(
            clip_txt_pooled=clip_text_pooled, clip_txt=clip_text, clip_img=clip_img
        )

        # Model Blocks
        x = self.embedding(x)
        if hasattr(self, "effnet_mapper") and effnet is not None:
            x = x + self.effnet_mapper(
                nn.functional.interpolate(
                    effnet, size=x.shape[-2:], mode="bilinear", align_corners=True
                )
            )
        if hasattr(self, "pixels_mapper"):
            x = x + nn.functional.interpolate(
                self.pixels_mapper(pixels),
                size=x.shape[-2:],
                mode="bilinear",
                align_corners=True,
            )
        level_outputs = self._down_encode(x, r_embed, clip)
        x = self.of_up_decode(level_outputs, r_embed, clip)

        return self.clf(x).to(torch.float16)


from contextlib import contextmanager

# diffusers.pipelines.stable_cascade.modeling_stable_cascade_common.StableCascadeUnet
from diffusers.pipelines.stable_cascade.modeling_stable_cascade_common import (
    StableCascadeUnet,
)

# torch2oflow_class_map.update({StableCascadeUnet: StableCascadeUnetOflow})
from onediff.infer_compiler.backends.oneflow.transform import register


@contextmanager
def patch_oneflow_prior_fp16_overflow():
    torch2oflow_class_map = {StableCascadeUnet: StableCascadeUnet_OF}
    register(torch2oflow_class_map=torch2oflow_class_map)
    yield
    torch2oflow_class_map = {StableCascadeUnet: StableCascadeUnet_OF_CLS}
    register(torch2oflow_class_map=torch2oflow_class_map)



try:
    import git  # type: ignore
except ImportError:
    print("GitPython library not found. Installing...")
    import subprocess

    subprocess.check_call(["pip", "install", "GitPython"])
    import git  # type: ignore

from datetime import datetime


class FunctionSelectorByCommitDate:
    def __init__(self, repo_path):
        self.repo_path = repo_path
        self.repo = git.Repo(repo_path)

    def _get_current_commit_date(self):
        """
        Get the date of the current commit.
        """
        current_commit = self.repo.head.commit
        current_commit_date = datetime.fromtimestamp(current_commit.committed_date)
        return current_commit, current_commit_date

    def _get_commit_date(self, commit_hash):
        """
        Get the date of a specific commit.
        """
        commit = self.repo.commit(commit_hash)
        commit_date = datetime.fromtimestamp(commit.committed_date)
        return commit_date

    def __call__(self, iterable, default_func=None):
        """
        Select a function based on the commit date.

        Args:
            iterable (iterable): An iterable of tuples, where each tuple contains
                a commit hash and a corresponding function.
            default_func (callable, optional): The default function to return
                if no match is found.

        Returns:
            callable: The selected function.
        """
        cur_commit, cur_date = self._get_current_commit_date()
        sel_date, sel_func = None, default_func
        for pair in iterable:
            commit_hash, func = pair
            other_commit_date = self._get_commit_date(commit_hash)
            if cur_date > other_commit_date and (
                sel_date and other_commit_date > sel_date
            ):
                sel_date = other_commit_date
                sel_func = func

        return sel_func



import argparse
import os
import time

import torch
import torch.nn as nn

# oneflow_compile should be imported before importing any diffusers
from onediff.infer_compiler import oneflow_compile, OneflowCompileOptions


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("--model", type=str, required=True)
    parser.add_argument(
        "--model_type",
        type=str,
        required=True,
        choices=["sd", "sdxl"],
        help="Specify the model type: 'sd' for Stable Diffusion or 'sdxl' for Stable Diffusion XL.",
    )
    parser.add_argument("--saved_image", type=str, required=True)
    parser.add_argument("--save_graph", action="store_true")
    parser.add_argument("--load_graph", action="store_true")
    parser.add_argument(
        "--prompt",
        type=str,
        default="photo of a rhino dressed suit and tie sitting at a table in a bar with a bar stools, award winning photography, Elke vogelsang",
    )
    parser.add_argument("--height", type=int, default=1024)
    parser.add_argument("--width", type=int, default=1024)
    parser.add_argument("--steps", type=int, default=30)
    parser.add_argument("--bits", type=int, default=8)
    parser.add_argument(
        "--compile",
        default=True,
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
    )
    parser.add_argument(
        "--compile_text_encoder",
        default=False,
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
        help=(
            "Switch controls whether text_encoder is compiled (default: False). "
            "If your CPU is powerful, turning it on will shorten end-to-end time."
        ),
    )
    parser.add_argument(
        "--graph",
        default=True,
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
    )
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument("--warmup", type=int, default=1)
    parser.add_argument("--cache_interval", type=int, default=3)
    parser.add_argument("--cache_layer_id", type=int, default=0)
    parser.add_argument("--cache_block_id", type=int, default=0)

    args = parser.parse_args()
    return args


args = parse_args()

assert os.path.isfile(
    os.path.join(args.model, "calibrate_info.txt")
), f"calibrate_info.txt is required in args.model ({args.model})"

if args.model_type == "sdxl":
    from onediffx.deep_cache import StableDiffusionXLPipeline
else:
    from onediffx.deep_cache import StableDiffusionPipeline
import onediff_quant
from onediff_quant.utils import replace_sub_module_with_quantizable_module

onediff_quant.enable_load_quantized_model()
infer_args = {
    "prompt": args.prompt,
    "height": args.height,
    "width": args.width,
    "num_inference_steps": args.steps,
    "cache_interval": args.cache_interval,
    "cache_layer_id": args.cache_layer_id,
    "cache_block_id": args.cache_block_id,
}
calibrate_info = {}
with open(os.path.join(args.model, "calibrate_info.txt"), "r") as f:
    for line in f.readlines():
        line = line.strip()
        items = line.split(" ")
        calibrate_info[items[0]] = [
            float(items[1]),
            int(items[2]),
            [float(x) for x in items[3].split(",")],
        ]

if args.model_type == "sdxl":
    pipe = StableDiffusionXLPipeline.from_pretrained(
        args.model,
        torch_dtype=torch.float16,
        use_safetensors=True,
        variant="fp16",
    )
else:
    pipe = StableDiffusionPipeline.from_pretrained(
        args.model,
        revision="fp16",
        variant="fp16",
        torch_dtype=torch.float16,
    )
pipe.to("cuda")

for sub_module_name, sub_calibrate_info in calibrate_info.items():
    replace_sub_module_with_quantizable_module(
        pipe.unet,
        sub_module_name,
        sub_calibrate_info,
        False,
        False,
        args.bits,
    )

compile_options = OneflowCompileOptions()
compile_options.use_graph = args.graph

if args.compile_text_encoder:
    if pipe.text_encoder is not None:
        pipe.text_encoder = oneflow_compile(pipe.text_encoder, options=compile_options)
    if hasattr(pipe, "text_encoder_2"):
        pipe.text_encoder_2 = oneflow_compile(
            pipe.text_encoder_2, options=compile_options
        )

if args.compile:
    if pipe.text_encoder is not None:
        pipe.text_encoder = oneflow_compile(pipe.text_encoder, options=compile_options)
    if hasattr(pipe, "text_encoder_2"):
        pipe.text_encoder_2 = oneflow_compile(
            pipe.text_encoder_2, options=compile_options
        )
    pipe.unet = oneflow_compile(pipe.unet, options=compile_options)
    pipe.fast_unet = oneflow_compile(pipe.fast_unet, options=compile_options)
    if hasattr(pipe, "text_encoder_2") and pipe.needs_upcasting:
        # To avoid mis-match of loaded graph and loaded model
        pipe.upcast_vae()
    pipe.vae.decoder = oneflow_compile(pipe.vae.decoder, options=compile_options)

if args.load_graph:
    print("Loading graphs to avoid compilation...")
    start_t = time.time()
    pipe.unet.load_graph("base_unet_compiled", run_warmup=True)
    pipe.fast_unet.load_graph("base_fast_unet_compiled", run_warmup=True)
    pipe.vae.decoder.load_graph("base_vae_compiled", run_warmup=True)
    end_t = time.time()
    print(f"warmup with loading graph elapsed: {end_t - start_t} s")
    start_t = time.time()
    for _ in range(args.warmup):
        torch.manual_seed(args.seed)
        image = pipe(**infer_args).images[0]
    end_t = time.time()
    print(f"warmup with run elapsed: {end_t - start_t} s")
else:
    start_t = time.time()
    for _ in range(args.warmup):
        torch.manual_seed(args.seed)
        image = pipe(**infer_args).images[0]
    end_t = time.time()
    print(f"warmup with run elapsed: {end_t - start_t} s")

start_t = time.time()

torch.manual_seed(args.seed)
torch.cuda.cudart().cudaProfilerStart()
image = pipe(**infer_args).images[0]
torch.cuda.cudart().cudaProfilerStop()

end_t = time.time()
print(f"e2e ({args.steps} steps) elapsed: {end_t - start_t} s")

image.save(args.saved_image)

if args.save_graph:
    print("Saving graphs...")
    start_t = time.time()
    pipe.unet.save_graph("base_unet_compiled")
    pipe.fast_unet.save_graph("base_fast_unet_compiled")
    pipe.vae.decoder.save_graph("base_vae_compiled")
    end_t = time.time()
    print(f"save graphs elapsed: {end_t - start_t} s")



import onediff_shared
import torch
from compile import is_oneflow_backend
from ldm.modules.diffusionmodules.openaimodel import UNetModel
from modules.sd_hijack_utils import CondFunc
from onediff_utils import check_structure_change, singleton_decorator

from .utils import get_controlnet_script


def hijacked_main_entry(self, p):
    self._original_controlnet_main_entry(p)
    from .compile import compile_controlnet_ldm_unet
    from .model import OnediffControlNetModel

    if not onediff_shared.onediff_enabled:
        return
    sd_ldm = p.sd_model
    unet = sd_ldm.model.diffusion_model

    structure_changed = check_structure_change(
        onediff_shared.previous_unet_type, sd_ldm
    )
    if not onediff_shared.controlnet_compiled or structure_changed:
        onediff_model = OnediffControlNetModel(unet)
        onediff_shared.current_unet_graph = compile_controlnet_ldm_unet(
            sd_ldm, onediff_model
        )
        onediff_shared.controlnet_compiled = True


# When OneDiff is initializing, the controlnet extension has not yet been loaded.
# Therefore, this function should be called during image generation
# rather than during the initialization of the OneDiff.
@singleton_decorator
def hijack_controlnet_extension(p):
    CondFunc(
        "scripts.hook.UnetHook.hook",
        hijacked_controlnet_hook,
        lambda _, *arg, **kwargs: onediff_shared.onediff_enabled,
    )
    # get controlnet script
    controlnet_script = get_controlnet_script(p)
    if controlnet_script is None:
        return

    controlnet_script._original_controlnet_main_entry = (
        controlnet_script.controlnet_main_entry
    )
    controlnet_script.controlnet_main_entry = hijacked_main_entry.__get__(
        controlnet_script
    )


def unhijack_controlnet_extension(p):
    controlnet_script = get_controlnet_script(p)
    if controlnet_script is None:
        return

    if hasattr(controlnet_script, "_original_controlnet_main_entry"):
        controlnet_script.controlnet_main_entry = (
            controlnet_script._original_controlnet_main_entry
        )
        delattr(controlnet_script, "_original_controlnet_main_entry")


# We were intended to only hack the closure function `forward`
# in the member function `hook` of the UnetHook class in the ControlNet extension.
# But due to certain limitations, we were unable to directly only hack
# the closure function `forward` within the `hook` method.
# So we have to hack the entire member function `hook` in the `UnetHook` class.

# The function largely retains its original content,
# with modifications specifically made within the `forward` function.
# To identify the altered parts, you can search for the tag "modified by OneDiff"

# https://github.com/Mikubill/sd-webui-controlnet/blob/8bbbd0e55ef6e5d71b09c2de2727b36e7bc825b0/scripts/hook.py#L442
def hijacked_controlnet_hook(
    orig_func,
    self,
    model,
    sd_ldm,
    control_params,
    process,
    batch_option_uint_separate=False,
    batch_option_style_align=False,
):
    from modules import devices, lowvram, scripts, shared
    from scripts.controlnet_sparsectrl import SparseCtrl
    from scripts.enums import AutoMachine, ControlModelType, HiResFixOption
    from scripts.hook import (
        AbstractLowScaleModel,
        blur,
        mark_prompt_context,
        predict_noise_from_start,
        predict_q_sample,
        predict_start_from_noise,
        register_schedule,
        torch_dfs,
        unmark_prompt_context,
    )
    from scripts.ipadapter.ipadapter_model import ImageEmbed
    from scripts.logging import logger

    self.model = model
    self.sd_ldm = sd_ldm
    self.control_params = control_params

    model_is_sdxl = getattr(self.sd_ldm, "is_sdxl", False)

    outer = self

    def process_sample(*args, **kwargs):
        # ControlNet must know whether a prompt is conditional prompt (positive prompt) or unconditional conditioning prompt (negative prompt).
        # You can use the hook.py's `mark_prompt_context` to mark the prompts that will be seen by ControlNet.
        # Let us say XXX is a MulticondLearnedConditioning or a ComposableScheduledPromptConditioning or a ScheduledPromptConditioning or a list of these components,
        # if XXX is a positive prompt, you should call mark_prompt_context(XXX, positive=True)
        # if XXX is a negative prompt, you should call mark_prompt_context(XXX, positive=False)
        # After you mark the prompts, the ControlNet will know which prompt is cond/uncond and works as expected.
        # After you mark the prompts, the mismatch errors will disappear.
        mark_prompt_context(kwargs.get("conditioning", []), positive=True)
        mark_prompt_context(
            kwargs.get("unconditional_conditioning", []), positive=False
        )
        mark_prompt_context(getattr(process, "hr_c", []), positive=True)
        mark_prompt_context(getattr(process, "hr_uc", []), positive=False)
        return process.sample_before_CN_hack(*args, **kwargs)

    def forward(self, x, timesteps=None, context=None, y=None, **kwargs):
        is_sdxl = y is not None and model_is_sdxl
        total_t2i_adapter_embedding = [0.0] * 4
        if is_sdxl:
            total_controlnet_embedding = [0.0] * 10
        else:
            total_controlnet_embedding = [0.0] * 13
        require_inpaint_hijack = False
        is_in_high_res_fix = False
        batch_size = int(x.shape[0])

        # Handle cond-uncond marker
        (
            cond_mark,
            outer.current_uc_indices,
            outer.current_c_indices,
            context,
        ) = unmark_prompt_context(context)
        outer.model.cond_mark = cond_mark
        # logger.info(str(cond_mark[:, 0, 0, 0].detach().cpu().numpy().tolist()) + ' - ' + str(outer.current_uc_indices))

        # Revision
        if is_sdxl:
            revision_y1280 = 0

            for param in outer.control_params:
                if param.guidance_stopped:
                    continue
                if param.control_model_type == ControlModelType.ReVision:
                    if param.vision_hint_count is None:
                        k = (
                            torch.Tensor(
                                [int(param.preprocessor["threshold_a"] * 1000)]
                            )
                            .to(param.hint_cond)
                            .long()
                            .clip(0, 999)
                        )
                        param.vision_hint_count = outer.revision_q_sampler.q_sample(
                            param.hint_cond, k
                        )
                    revision_emb = param.vision_hint_count
                    if isinstance(revision_emb, torch.Tensor):
                        revision_y1280 += revision_emb * param.weight

            if isinstance(revision_y1280, torch.Tensor):
                y[:, :1280] = revision_y1280 * cond_mark[:, :, 0, 0]
                if any(
                    "ignore_prompt" in param.preprocessor["name"]
                    for param in outer.control_params
                ) or (
                    getattr(process, "prompt", "") == ""
                    and getattr(process, "negative_prompt", "") == ""
                ):
                    context = torch.zeros_like(context)

        # High-res fix
        for param in outer.control_params:
            # select which hint_cond to use
            if param.used_hint_cond is None:
                param.used_hint_cond = param.hint_cond
                param.used_hint_cond_latent = None
                param.used_hint_inpaint_hijack = None

            # has high-res fix
            if (
                isinstance(param.hr_hint_cond, torch.Tensor)
                and x.ndim == 4
                and param.hint_cond.ndim == 4
                and param.hr_hint_cond.ndim == 4
            ):
                _, _, h_lr, w_lr = param.hint_cond.shape
                _, _, h_hr, w_hr = param.hr_hint_cond.shape
                _, _, h, w = x.shape
                h, w = h * 8, w * 8
                if abs(h - h_lr) < abs(h - h_hr):
                    is_in_high_res_fix = False
                    if param.used_hint_cond is not param.hint_cond:
                        param.used_hint_cond = param.hint_cond
                        param.used_hint_cond_latent = None
                        param.used_hint_inpaint_hijack = None
                else:
                    is_in_high_res_fix = True
                    if param.used_hint_cond is not param.hr_hint_cond:
                        param.used_hint_cond = param.hr_hint_cond
                        param.used_hint_cond_latent = None
                        param.used_hint_inpaint_hijack = None

        self.is_in_high_res_fix = is_in_high_res_fix
        outer.is_in_high_res_fix = is_in_high_res_fix

        # Convert control image to latent
        for param in outer.control_params:
            if param.used_hint_cond_latent is not None:
                continue
            if (
                param.control_model_type not in [ControlModelType.AttentionInjection]
                and "colorfix" not in param.preprocessor["name"]
                and "inpaint_only" not in param.preprocessor["name"]
            ):
                continue
            param.used_hint_cond_latent = outer.call_vae_using_process(
                process, param.used_hint_cond, batch_size=batch_size
            )

        # vram
        for param in outer.control_params:
            if getattr(param.control_model, "disable_memory_management", False):
                continue

            if param.control_model is not None:
                if (
                    outer.lowvram
                    and is_sdxl
                    and hasattr(param.control_model, "aggressive_lowvram")
                ):
                    param.control_model.aggressive_lowvram()
                elif hasattr(param.control_model, "fullvram"):
                    param.control_model.fullvram()
                elif hasattr(param.control_model, "to"):
                    param.control_model.to(devices.get_device_for("controlnet"))

        # handle prompt token control
        for param in outer.control_params:
            if param.guidance_stopped or param.disabled_by_hr_option(
                self.is_in_high_res_fix
            ):
                continue

            if param.control_model_type not in [ControlModelType.T2I_StyleAdapter]:
                continue

            control = param.control_model(
                x=x, hint=param.used_hint_cond, timesteps=timesteps, context=context
            )
            control = torch.cat([control.clone() for _ in range(batch_size)], dim=0)
            control *= param.weight
            control *= cond_mark[:, :, :, 0]
            context = torch.cat([context, control.clone()], dim=1)

        # handle ControlNet / T2I_Adapter
        for param_index, param in enumerate(outer.control_params):
            if param.guidance_stopped or param.disabled_by_hr_option(
                self.is_in_high_res_fix
            ):
                continue

            if not (
                param.control_model_type.is_controlnet
                or param.control_model_type == ControlModelType.T2I_Adapter
            ):
                continue

            # inpaint model workaround
            x_in = x
            control_model = param.control_model.control_model

            if param.control_model_type.is_controlnet:
                if (
                    x.shape[1] != control_model.input_blocks[0][0].in_channels
                    and x.shape[1] == 9
                ):
                    # inpaint_model: 4 data + 4 downscaled image + 1 mask
                    x_in = x[:, :4, ...]
                    require_inpaint_hijack = True

            assert (
                param.used_hint_cond is not None
            ), "Controlnet is enabled but no input image is given"

            hint = param.used_hint_cond
            if param.control_model_type == ControlModelType.InstantID:
                assert isinstance(param.control_context_override, ImageEmbed)
                controlnet_context = param.control_context_override.eval(cond_mark).to(
                    x.device, dtype=x.dtype
                )
            else:
                controlnet_context = context

            # ControlNet inpaint protocol
            if hint.shape[1] == 4 and not isinstance(control_model, SparseCtrl):
                c = hint[:, 0:3, :, :]
                m = hint[:, 3:4, :, :]
                m = (m > 0.5).float()
                hint = c * (1 - m) - m

            control = param.control_model(
                x=x_in, hint=hint, timesteps=timesteps, context=controlnet_context, y=y
            )

            if is_sdxl:
                control_scales = [param.weight] * 10
            else:
                control_scales = [param.weight] * 13

            if param.cfg_injection or param.global_average_pooling:
                if param.control_model_type == ControlModelType.T2I_Adapter:
                    control = [
                        torch.cat([c.clone() for _ in range(batch_size)], dim=0)
                        for c in control
                    ]
                control = [c * cond_mark for c in control]

            high_res_fix_forced_soft_injection = False

            if is_in_high_res_fix:
                if "canny" in param.preprocessor["name"]:
                    high_res_fix_forced_soft_injection = True
                if "mlsd" in param.preprocessor["name"]:
                    high_res_fix_forced_soft_injection = True

            if param.soft_injection or high_res_fix_forced_soft_injection:
                # important! use the soft weights with high-res fix can significantly reduce artifacts.
                if param.control_model_type == ControlModelType.T2I_Adapter:
                    control_scales = [
                        param.weight * x for x in (0.25, 0.62, 0.825, 1.0)
                    ]
                elif param.control_model_type.is_controlnet:
                    control_scales = [
                        param.weight * (0.825 ** float(12 - i)) for i in range(13)
                    ]

            if is_sdxl and param.control_model_type.is_controlnet:
                control_scales = control_scales[:10]

            if param.advanced_weighting is not None:
                logger.info(f"Advanced weighting enabled. {param.advanced_weighting}")
                if param.soft_injection or high_res_fix_forced_soft_injection:
                    logger.warn("Advanced weighting overwrites soft_injection effect.")
                control_scales = param.advanced_weighting

            control = [
                param.apply_effective_region_mask(c * scale)
                for c, scale in zip(control, control_scales)
            ]
            if param.global_average_pooling:
                control = [torch.mean(c, dim=(2, 3), keepdim=True) for c in control]

            for idx, item in enumerate(control):
                target = None
                if param.control_model_type.is_controlnet:
                    target = total_controlnet_embedding
                if param.control_model_type == ControlModelType.T2I_Adapter:
                    target = total_t2i_adapter_embedding
                if target is not None:
                    if batch_option_uint_separate:
                        for pi, ci in enumerate(outer.current_c_indices):
                            if pi % len(outer.control_params) != param_index:
                                item[ci] = 0
                        for pi, ci in enumerate(outer.current_uc_indices):
                            if pi % len(outer.control_params) != param_index:
                                item[ci] = 0
                        target[idx] = item + target[idx]
                    else:
                        target[idx] = item + target[idx]

        # Replace x_t to support inpaint models
        for param in outer.control_params:
            if not isinstance(param.used_hint_cond, torch.Tensor):
                continue
            if param.used_hint_cond.ndim < 2 or param.used_hint_cond.shape[1] != 4:
                continue
            if x.shape[1] != 9:
                continue
            if param.used_hint_inpaint_hijack is None:
                mask_pixel = param.used_hint_cond[:, 3:4, :, :]
                image_pixel = param.used_hint_cond[:, 0:3, :, :]
                mask_pixel = (mask_pixel > 0.5).to(mask_pixel.dtype)
                masked_latent = outer.call_vae_using_process(
                    process, image_pixel, batch_size, mask=mask_pixel
                )
                mask_latent = torch.nn.functional.max_pool2d(mask_pixel, (8, 8))
                if mask_latent.shape[0] != batch_size:
                    mask_latent = torch.cat(
                        [mask_latent.clone() for _ in range(batch_size)], dim=0
                    )
                param.used_hint_inpaint_hijack = torch.cat(
                    [mask_latent, masked_latent], dim=1
                )
                param.used_hint_inpaint_hijack.to(x.dtype).to(x.device)
            x = torch.cat([x[:, :4, :, :], param.used_hint_inpaint_hijack], dim=1)

        # vram
        for param in outer.control_params:
            if param.control_model is not None:
                if outer.lowvram:
                    param.control_model.to("cpu")

        # A1111 fix for medvram.
        if shared.cmd_opts.medvram or (
            getattr(shared.cmd_opts, "medvram_sdxl", False) and is_sdxl
        ):
            try:
                # Trigger the register_forward_pre_hook
                outer.sd_ldm.model()
            except Exception as e:
                logger.debug("register_forward_pre_hook")
                logger.debug(e)

        # Clear attention and AdaIn cache
        for module in outer.attn_module_list:
            module.bank = []
            module.style_cfgs = []
        for module in outer.gn_module_list:
            module.mean_bank = []
            module.var_bank = []
            module.style_cfgs = []

        # Handle attention and AdaIn control
        for param in outer.control_params:
            if param.guidance_stopped or param.disabled_by_hr_option(
                self.is_in_high_res_fix
            ):
                continue

            if param.used_hint_cond_latent is None:
                continue

            if param.control_model_type not in [ControlModelType.AttentionInjection]:
                continue

            ref_xt = predict_q_sample(
                outer.sd_ldm,
                param.used_hint_cond_latent,
                torch.round(timesteps.float()).long(),
            )

            # Inpaint Hijack
            if x.shape[1] == 9:
                ref_xt = torch.cat(
                    [
                        ref_xt,
                        torch.zeros_like(ref_xt)[:, 0:1, :, :],
                        param.used_hint_cond_latent,
                    ],
                    dim=1,
                )

            outer.current_style_fidelity = float(param.preprocessor["threshold_a"])
            outer.current_style_fidelity = max(
                0.0, min(1.0, outer.current_style_fidelity)
            )

            if is_sdxl:
                # sdxl's attention hacking is highly unstable.
                # We have no other methods but to reduce the style_fidelity a bit.
                # By default, 0.5 ** 3.0 = 0.125
                outer.current_style_fidelity = outer.current_style_fidelity**3.0

            if param.cfg_injection:
                outer.current_style_fidelity = 1.0
            elif param.soft_injection or is_in_high_res_fix:
                outer.current_style_fidelity = 0.0

            control_name = param.preprocessor["name"]

            if control_name in ["reference_only", "reference_adain+attn"]:
                outer.attention_auto_machine = AutoMachine.Write
                outer.attention_auto_machine_weight = param.weight

            if control_name in ["reference_adain", "reference_adain+attn"]:
                outer.gn_auto_machine = AutoMachine.Write
                outer.gn_auto_machine_weight = param.weight

            if is_sdxl:
                outer.original_forward(
                    x=ref_xt.to(devices.dtype_unet),
                    timesteps=timesteps.to(devices.dtype_unet),
                    context=context.to(devices.dtype_unet),
                    y=y,
                )
            else:
                outer.original_forward(
                    x=ref_xt.to(devices.dtype_unet),
                    timesteps=timesteps.to(devices.dtype_unet),
                    context=context.to(devices.dtype_unet),
                )

            outer.attention_auto_machine = AutoMachine.Read
            outer.gn_auto_machine = AutoMachine.Read

        # ------ modified by OneDiff below ------
        x = x.half()
        context = context.half()
        y = y.half() if y is not None else y
        h = onediff_shared.current_unet_graph.graph_module(
            x,
            timesteps,
            context,
            y,
            total_t2i_adapter_embedding,
            total_controlnet_embedding,
            is_sdxl,
            require_inpaint_hijack,
        )
        # ------ modified by OneDiff above ------

        # Post-processing for color fix
        for param in outer.control_params:
            if param.used_hint_cond_latent is None:
                continue
            if "colorfix" not in param.preprocessor["name"]:
                continue

            k = int(param.preprocessor["threshold_a"])
            if is_in_high_res_fix and not param.disabled_by_hr_option(
                self.is_in_high_res_fix
            ):
                k *= 2

            # Inpaint hijack
            xt = x[:, :4, :, :]

            x0_origin = param.used_hint_cond_latent
            t = torch.round(timesteps.float()).long()
            x0_prd = predict_start_from_noise(outer.sd_ldm, xt, t, h)
            x0 = x0_prd - blur(x0_prd, k) + blur(x0_origin, k)

            if "+sharp" in param.preprocessor["name"]:
                detail_weight = float(param.preprocessor["threshold_b"]) * 0.01
                neg = detail_weight * blur(x0, k) + (1 - detail_weight) * x0
                x0 = cond_mark * x0 + (1 - cond_mark) * neg

            eps_prd = predict_noise_from_start(outer.sd_ldm, xt, t, x0)

            w = max(0.0, min(1.0, float(param.weight)))
            h = eps_prd * w + h * (1 - w)

        # Post-processing for restore
        for param in outer.control_params:
            if param.used_hint_cond_latent is None:
                continue
            if "inpaint_only" not in param.preprocessor["name"]:
                continue
            if param.used_hint_cond.shape[1] != 4:
                continue

            # Inpaint hijack
            xt = x[:, :4, :, :]

            mask = param.used_hint_cond[:, 3:4, :, :]
            mask = torch.nn.functional.max_pool2d(
                mask, (10, 10), stride=(8, 8), padding=1
            )

            x0_origin = param.used_hint_cond_latent
            t = torch.round(timesteps.float()).long()
            x0_prd = predict_start_from_noise(outer.sd_ldm, xt, t, h)
            x0 = x0_prd * mask + x0_origin * (1 - mask)
            eps_prd = predict_noise_from_start(outer.sd_ldm, xt, t, x0)

            w = max(0.0, min(1.0, float(param.weight)))
            h = eps_prd * w + h * (1 - w)

        return h

    def move_all_control_model_to_cpu():
        for param in getattr(outer, "control_params", []) or []:
            if isinstance(param.control_model, torch.nn.Module):
                param.control_model.to("cpu")

    def forward_webui(*args, **kwargs):
        # ------ modified by OneDiff below ------
        forward_func = None
        graph_module = onediff_shared.current_unet_graph.graph_module
        if is_oneflow_backend():
            if "forward" in graph_module._torch_module.__dict__:
                forward_func = graph_module._torch_module.__dict__.pop("forward")
                _original_forward_func = graph_module._torch_module.__dict__.pop(
                    "_original_forward"
                )
        # ------ modified by OneDiff above ------

        # webui will handle other compoments
        try:
            if shared.cmd_opts.lowvram:
                lowvram.send_everything_to_cpu()
            return forward(*args, **kwargs)
        except Exception as e:
            move_all_control_model_to_cpu()
            raise e
        finally:
            if outer.lowvram:
                move_all_control_model_to_cpu()

            # ------ modified by OneDiff below ------
            if is_oneflow_backend():
                if forward_func is not None:
                    graph_module._torch_module.forward = forward_func
                    graph_module._torch_module._original_forward = (
                        _original_forward_func
                    )
            # ------ modified by OneDiff above ------

    def hacked_basic_transformer_inner_forward(self, x, context=None):
        x_norm1 = self.norm1(x)
        self_attn1 = None
        if self.disable_self_attn:
            # Do not use self-attention
            self_attn1 = self.attn1(x_norm1, context=context)
        else:
            # Use self-attention
            self_attention_context = x_norm1
            if outer.attention_auto_machine == AutoMachine.Write:
                if outer.attention_auto_machine_weight > self.attn_weight:
                    self.bank.append(self_attention_context.detach().clone())
                    self.style_cfgs.append(outer.current_style_fidelity)
            if outer.attention_auto_machine == AutoMachine.Read:
                if len(self.bank) > 0:
                    style_cfg = sum(self.style_cfgs) / float(len(self.style_cfgs))
                    self_attn1_uc = self.attn1(
                        x_norm1,
                        context=torch.cat([self_attention_context] + self.bank, dim=1),
                    )
                    self_attn1_c = self_attn1_uc.clone()
                    if len(outer.current_uc_indices) > 0 and style_cfg > 1e-5:
                        self_attn1_c[outer.current_uc_indices] = self.attn1(
                            x_norm1[outer.current_uc_indices],
                            context=self_attention_context[outer.current_uc_indices],
                        )
                    self_attn1 = (
                        style_cfg * self_attn1_c + (1.0 - style_cfg) * self_attn1_uc
                    )
                self.bank = []
                self.style_cfgs = []
            if (
                outer.attention_auto_machine == AutoMachine.StyleAlign
                and not outer.is_in_high_res_fix
            ):
                # very VRAM hungry - disable at high_res_fix

                def shared_attn1(inner_x):
                    BB, FF, CC = inner_x.shape
                    return self.attn1(inner_x.reshape(1, BB * FF, CC)).reshape(
                        BB, FF, CC
                    )

                uc_layer = shared_attn1(x_norm1[outer.current_uc_indices])
                c_layer = shared_attn1(x_norm1[outer.current_c_indices])
                self_attn1 = torch.zeros_like(x_norm1).to(uc_layer)
                self_attn1[outer.current_uc_indices] = uc_layer
                self_attn1[outer.current_c_indices] = c_layer
                del uc_layer, c_layer
            if self_attn1 is None:
                self_attn1 = self.attn1(x_norm1, context=self_attention_context)

        x = self_attn1.to(x.dtype) + x
        x = self.attn2(self.norm2(x), context=context) + x
        x = self.ff(self.norm3(x)) + x
        return x

    def hacked_group_norm_forward(self, *args, **kwargs):
        eps = 1e-6
        x = self.original_forward_cn_hijack(*args, **kwargs)
        y = None
        if outer.gn_auto_machine == AutoMachine.Write:
            if outer.gn_auto_machine_weight > self.gn_weight:
                var, mean = torch.var_mean(x, dim=(2, 3), keepdim=True, correction=0)
                self.mean_bank.append(mean)
                self.var_bank.append(var)
                self.style_cfgs.append(outer.current_style_fidelity)
        if outer.gn_auto_machine == AutoMachine.Read:
            if len(self.mean_bank) > 0 and len(self.var_bank) > 0:
                style_cfg = sum(self.style_cfgs) / float(len(self.style_cfgs))
                var, mean = torch.var_mean(x, dim=(2, 3), keepdim=True, correction=0)
                std = torch.maximum(var, torch.zeros_like(var) + eps) ** 0.5
                mean_acc = sum(self.mean_bank) / float(len(self.mean_bank))
                var_acc = sum(self.var_bank) / float(len(self.var_bank))
                std_acc = torch.maximum(var_acc, torch.zeros_like(var_acc) + eps) ** 0.5
                y_uc = (((x - mean) / std) * std_acc) + mean_acc
                y_c = y_uc.clone()
                if len(outer.current_uc_indices) > 0 and style_cfg > 1e-5:
                    y_c[outer.current_uc_indices] = x.to(y_c.dtype)[
                        outer.current_uc_indices
                    ]
                y = style_cfg * y_c + (1.0 - style_cfg) * y_uc
            self.mean_bank = []
            self.var_bank = []
            self.style_cfgs = []
        if y is None:
            y = x
        return y.to(x.dtype)

    if getattr(process, "sample_before_CN_hack", None) is None:
        process.sample_before_CN_hack = process.sample
    process.sample = process_sample

    model._original_forward = model.forward
    outer.original_forward = model.forward
    model.forward = forward_webui.__get__(model, UNetModel)

    if model_is_sdxl:
        register_schedule(sd_ldm)
        outer.revision_q_sampler = AbstractLowScaleModel()

    need_attention_hijack = False

    for param in outer.control_params:
        if param.control_model_type in [ControlModelType.AttentionInjection]:
            need_attention_hijack = True

    if batch_option_style_align:
        need_attention_hijack = True
        outer.attention_auto_machine = AutoMachine.StyleAlign
        outer.gn_auto_machine = AutoMachine.StyleAlign

    all_modules = torch_dfs(model)

    if need_attention_hijack:
        attn_modules = [
            module
            for module in all_modules
            if isinstance(module, BasicTransformerBlock)
            or isinstance(module, BasicTransformerBlockSGM)
        ]
        attn_modules = sorted(attn_modules, key=lambda x: -x.norm1.normalized_shape[0])

        for i, module in enumerate(attn_modules):
            if getattr(module, "_original_inner_forward_cn_hijack", None) is None:
                module._original_inner_forward_cn_hijack = module._forward
            module._forward = hacked_basic_transformer_inner_forward.__get__(
                module, BasicTransformerBlock
            )
            module.bank = []
            module.style_cfgs = []
            module.attn_weight = float(i) / float(len(attn_modules))

        gn_modules = [model.middle_block]
        model.middle_block.gn_weight = 0

        if model_is_sdxl:
            input_block_indices = [4, 5, 7, 8]
            output_block_indices = [0, 1, 2, 3, 4, 5]
        else:
            input_block_indices = [4, 5, 7, 8, 10, 11]
            output_block_indices = [0, 1, 2, 3, 4, 5, 6, 7]

        for w, i in enumerate(input_block_indices):
            module = model.input_blocks[i]
            module.gn_weight = 1.0 - float(w) / float(len(input_block_indices))
            gn_modules.append(module)

        for w, i in enumerate(output_block_indices):
            module = model.output_blocks[i]
            module.gn_weight = float(w) / float(len(output_block_indices))
            gn_modules.append(module)

        for i, module in enumerate(gn_modules):
            if getattr(module, "original_forward_cn_hijack", None) is None:
                module.original_forward_cn_hijack = module.forward
            module.forward = hacked_group_norm_forward.__get__(module, torch.nn.Module)
            module.mean_bank = []
            module.var_bank = []
            module.style_cfgs = []
            module.gn_weight *= 2

        outer.attn_module_list = attn_modules
        outer.gn_module_list = gn_modules
    else:
        for module in all_modules:
            _original_inner_forward_cn_hijack = getattr(
                module, "_original_inner_forward_cn_hijack", None
            )
            original_forward_cn_hijack = getattr(
                module, "original_forward_cn_hijack", None
            )
            if _original_inner_forward_cn_hijack is not None:
                module._forward = _original_inner_forward_cn_hijack
            if original_forward_cn_hijack is not None:
                module.forward = original_forward_cn_hijack
        outer.attn_module_list = []
        outer.gn_module_list = []

    scripts.script_callbacks.on_cfg_denoiser(self.guidance_schedule_handler)



from pathlib import Path
from typing import Dict, Union

from modules.sd_models import select_checkpoint

from onediff.optimization.quant_optimizer import (
    quantize_model,
    varify_can_use_quantization,
)
from onediff.utils import logger


def quant_unet_oneflow(compiled_unet):
    if varify_can_use_quantization():
        calibrate_info = get_calibrate_info(
            f"{Path(select_checkpoint().filename).stem}_sd_calibrate_info.txt"
        )
        compiled_unet = quantize_model(
            compiled_unet, inplace=False, calibrate_info=calibrate_info
        )
    return compiled_unet


def get_calibrate_info(filename: str) -> Union[None, Dict]:
    calibration_path = Path(select_checkpoint().filename).parent / filename
    if not calibration_path.exists():
        return None

    logger.info(f"Got calibrate info at {str(calibration_path)}")
    calibrate_info = {}
    with open(calibration_path, "r") as f:
        for line in f.readlines():
            line = line.strip()
            items = line.split(" ")
            calibrate_info[items[0]] = [
                float(items[1]),
                int(items[2]),
                [float(x) for x in items[3].split(",")],
            ]
    return calibrate_info



from onediff.infer_compiler.backends.oneflow.param_utils import (
    update_graph_with_constant_folding_info,
)

from .lora import (
    delete_adapters,
    get_active_adapters,
    load_and_fuse_lora,
    load_lora_and_optionally_fuse,
    set_and_fuse_adapters,
    unfuse_lora,
)

__all__ = [
    "delete_adapters",
    "get_active_adapters",
    "load_and_fuse_lora",
    "load_lora_and_optionally_fuse",
    "set_and_fuse_adapters",
    "unfuse_lora",
]



import comfy
import comfy.utils
import torch
import torchvision.transforms as T
from facexlib.parsing import init_parsing_model
from facexlib.utils.face_restoration_helper import FaceRestoreHelper
from register_comfy.CrossAttentionPatch import pulid_attention

from ..hijack_ipadapter_plus.set_model_patch_replace import apply_patch
from ..utils.booster_utils import is_using_oneflow_backend
from ._config import pulid_comfyui_hijacker, pulid_comfyui_pt

pulid_pkg = pulid_comfyui_pt.pulid
PulidModel = pulid_pkg.PulidModel
tensor_to_image = pulid_pkg.tensor_to_image
image_to_tensor = pulid_pkg.image_to_tensor
to_gray = pulid_pkg.to_gray


def set_model_patch_replace(model, patch_kwargs, key):
    apply_patch(
        org_fn=None,
        model=model,
        patch_kwargs=patch_kwargs,
        key=key,
        attention_func=pulid_attention,
    )


def apply_pulid(
    org_fn,
    self,
    model,
    pulid,
    eva_clip,
    face_analysis,
    image,
    weight,
    start_at,
    end_at,
    method=None,
    noise=0.0,
    fidelity=None,
    projection=None,
    attn_mask=None,
):
    work_model = model.clone()

    device = comfy.model_management.get_torch_device()
    dtype = comfy.model_management.unet_dtype()
    if dtype not in [torch.float32, torch.float16, torch.bfloat16]:
        dtype = (
            torch.float16 if comfy.model_management.should_use_fp16() else torch.float32
        )

    eva_clip.to(device, dtype=dtype)
    # pulid_model = PulidModel(pulid).to(device, dtype=dtype)
    if hasattr(self, "_pulid_model"):
        cached_pulid, pulid_model = self._pulid_model
        if cached_pulid == pulid:
            print(f"Model is cached.")
        else:
            print(f"Cached model not found for pulid: {pulid}.")
            pulid_model = PulidModel(pulid).to(device, dtype=dtype)
            self._pulid_model = (pulid, pulid_model)
    else:
        pulid_model = PulidModel(pulid).to(device, dtype=dtype)
        self._pulid_model = (pulid, pulid_model)
        print(f"Cached Model")

    if attn_mask is not None:
        if attn_mask.dim() > 3:
            attn_mask = attn_mask.squeeze(-1)
        elif attn_mask.dim() < 3:
            attn_mask = attn_mask.unsqueeze(0)
        attn_mask = attn_mask.to(device, dtype=dtype)

    if method == "fidelity" or projection == "ortho_v2":
        num_zero = 8
        ortho = False
        ortho_v2 = True
    elif method == "style" or projection == "ortho":
        num_zero = 16
        ortho = True
        ortho_v2 = False
    else:
        num_zero = 0
        ortho = False
        ortho_v2 = False

    if fidelity is not None:
        num_zero = fidelity

    # face_analysis.det_model.input_size = (640,640)
    image = tensor_to_image(image)

    if hasattr(self, "_face_helper_info"):
        face_helper = self._face_helper_info
    else:
        face_helper = FaceRestoreHelper(
            upscale_factor=1,
            face_size=512,
            crop_ratio=(1, 1),
            det_model="retinaface_resnet50",
            save_ext="png",
            device=device,
        )

        face_helper.face_parse = None
        face_helper.face_parse = init_parsing_model(model_name="bisenet", device=device)
        self._face_helper_info = face_helper

    bg_label = [0, 16, 18, 7, 8, 9, 14, 15]
    cond = []
    uncond = []

    for i in range(image.shape[0]):
        # get insightface embeddings
        iface_embeds = None
        for size in [(size, size) for size in range(640, 256, -64)]:
            face_analysis.det_model.input_size = size
            face = face_analysis.get(image[i])
            if face:
                face = sorted(
                    face,
                    key=lambda x: (x.bbox[2] - x.bbox[0]) * (x.bbox[3] - x.bbox[1]),
                    reverse=True,
                )[-1]
                iface_embeds = (
                    torch.from_numpy(face.embedding)
                    .unsqueeze(0)
                    .to(device, dtype=dtype)
                )
                break
        else:
            raise Exception("insightface: No face detected.")

        # get eva_clip embeddings
        face_helper.clean_all()
        face_helper.read_image(image[i])
        face_helper.get_face_landmarks_5(only_center_face=True)
        face_helper.align_warp_face()

        if len(face_helper.cropped_faces) == 0:
            raise Exception("facexlib: No face detected.")

        face = face_helper.cropped_faces[0]
        face = image_to_tensor(face).unsqueeze(0).permute(0, 3, 1, 2).to(device)
        parsing_out = face_helper.face_parse(
            T.functional.normalize(face, [0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        )[0]
        parsing_out = parsing_out.argmax(dim=1, keepdim=True)
        bg = sum(parsing_out == i for i in bg_label).bool()
        white_image = torch.ones_like(face)
        face_features_image = torch.where(bg, white_image, to_gray(face))
        face_features_image = T.functional.resize(
            face_features_image,
            eva_clip.image_size,
            T.InterpolationMode.BICUBIC
            if "cuda" in device.type
            else T.InterpolationMode.BILINEAR,
        ).to(device, dtype=dtype)
        face_features_image = T.functional.normalize(
            face_features_image, eva_clip.image_mean, eva_clip.image_std
        )

        id_cond_vit, id_vit_hidden = eva_clip(
            face_features_image,
            return_all_features=False,
            return_hidden=True,
            shuffle=False,
        )
        id_cond_vit = id_cond_vit.to(device, dtype=dtype)
        for idx in range(len(id_vit_hidden)):
            id_vit_hidden[idx] = id_vit_hidden[idx].to(device, dtype=dtype)

        id_cond_vit = torch.div(id_cond_vit, torch.norm(id_cond_vit, 2, 1, True))

        # combine embeddings
        id_cond = torch.cat([iface_embeds, id_cond_vit], dim=-1)
        if noise == 0:
            id_uncond = torch.zeros_like(id_cond)
        else:
            id_uncond = torch.rand_like(id_cond) * noise
        id_vit_hidden_uncond = []
        for idx in range(len(id_vit_hidden)):
            if noise == 0:
                id_vit_hidden_uncond.append(torch.zeros_like(id_vit_hidden[idx]))
            else:
                id_vit_hidden_uncond.append(torch.rand_like(id_vit_hidden[idx]) * noise)

        cond.append(pulid_model.get_image_embeds(id_cond, id_vit_hidden))
        uncond.append(pulid_model.get_image_embeds(id_uncond, id_vit_hidden_uncond))

    # average embeddings
    cond = torch.cat(cond).to(device, dtype=dtype)
    uncond = torch.cat(uncond).to(device, dtype=dtype)
    if cond.shape[0] > 1:
        cond = torch.mean(cond, dim=0, keepdim=True)
        uncond = torch.mean(uncond, dim=0, keepdim=True)

    if num_zero > 0:
        if noise == 0:
            zero_tensor = torch.zeros(
                (cond.size(0), num_zero, cond.size(-1)), dtype=dtype, device=device
            )
        else:
            zero_tensor = (
                torch.rand(
                    (cond.size(0), num_zero, cond.size(-1)), dtype=dtype, device=device
                )
                * noise
            )
        cond = torch.cat([cond, zero_tensor], dim=1)
        uncond = torch.cat([uncond, zero_tensor], dim=1)

    sigma_start = work_model.get_model_object("model_sampling").percent_to_sigma(
        start_at
    )
    sigma_end = work_model.get_model_object("model_sampling").percent_to_sigma(end_at)

    patch_kwargs = {
        "pulid": pulid_model,
        "weight": weight,
        "cond": cond,
        "uncond": uncond,
        "sigma_start": sigma_start,
        "sigma_end": sigma_end,
        "ortho": ortho,
        "ortho_v2": ortho_v2,
        "mask": attn_mask,
    }

    number = 0
    for id in [4, 5, 7, 8]:  # id of input_blocks that have cross attention
        block_indices = range(2) if id in [4, 5] else range(10)  # transformer_depth
        for index in block_indices:
            patch_kwargs["module_key"] = str(number * 2 + 1)
            set_model_patch_replace(work_model, patch_kwargs, ("input", id, index))
            number += 1
    for id in range(6):  # id of output_blocks that have cross attention
        block_indices = range(2) if id in [3, 4, 5] else range(10)  # transformer_depth
        for index in block_indices:
            patch_kwargs["module_key"] = str(number * 2 + 1)
            set_model_patch_replace(work_model, patch_kwargs, ("output", id, index))
            number += 1
    for index in range(10):
        patch_kwargs["module_key"] = str(number * 2 + 1)
        set_model_patch_replace(work_model, patch_kwargs, ("middle", 0, index))
        number += 1

    return (work_model,)


def apply_pulid_cond_func(org_fn, self, model, *args, **kwargs):
    return is_using_oneflow_backend(model)


pulid_comfyui_hijacker.register(
    pulid_comfyui_pt.pulid.ApplyPulid.apply_pulid, apply_pulid, apply_pulid_cond_func
)



import copy

import comfy
import torch

from ..infer_compiler_registry.register_comfy import DeepCacheUNet, FastDeepCacheUNet


def state_dict_hook(module, state_dict, prefix, local_metadata):
    new_state_dict = type(state_dict)()
    for k, v in state_dict.items():
        # diffusion_model._deployable_module_model._torch_module.out.2.weight => diffusion_model.out.2.weight
        if k.startswith("diffusion_model._deployable_module_model"):
            x = k.split(".")
            new_k = ".".join(x[:1] + x[3:])
            new_state_dict[new_k] = v
        else:
            new_state_dict[k] = v
    return new_state_dict


class OneFlowSpeedUpModelPatcher(comfy.model_patcher.ModelPatcher):
    def __init__(
        self,
        model,
        load_device,
        offload_device,
        size=0,
        current_device=None,
        weight_inplace_update=False,
        *,
        use_graph=None,
        graph_path=None,
        graph_device=None,
    ):
        from onediff.infer_compiler import (
            DeployableModule,
            oneflow_compile,
            OneflowCompileOptions,
        )

        self.weight_inplace_update = weight_inplace_update
        self.object_patches = {}
        self.object_patches_backup = {}
        self.size = size
        self.model = copy.copy(model)
        self.model.__dict__["_modules"] = copy.copy(model.__dict__["_modules"])
        if isinstance(self.model.diffusion_model, DeployableModule):
            self.model.__dict__["_modules"][
                "diffusion_model"
            ] = self.model.diffusion_model
        else:
            options = OneflowCompileOptions()
            options.use_graph = use_graph
            options.graph_file = graph_path
            options.graph_file_device = graph_device
            self.model.__dict__["_modules"]["diffusion_model"] = oneflow_compile(
                self.model.diffusion_model, options=options
            )
        self.model._register_state_dict_hook(state_dict_hook)
        self.patches = {}
        self.backup = {}
        self.model_options = {"transformer_options": {}}
        self.model_size()
        self.load_device = load_device
        self.offload_device = offload_device
        if current_device is None:
            self.current_device = self.offload_device
        else:
            self.current_device = current_device

        self.model_lowvram = False

    def clone(self):
        n = OneFlowSpeedUpModelPatcher(
            self.model,
            self.load_device,
            self.offload_device,
            self.size,
            self.current_device,
            weight_inplace_update=self.weight_inplace_update,
        )
        n.patches = {}
        for k in self.patches:
            n.patches[k] = self.patches[k][:]

        n.object_patches = self.object_patches.copy()
        n.model_options = copy.deepcopy(self.model_options)
        n.model_keys = self.model_keys
        self.model_lowvram = False
        return n

    def add_patches(self, patches, strength_patch=1.0, strength_model=1.0):
        from comfy.ldm.modules.attention import CrossAttention

        is_onediff_quant_available = False
        try:
            import onediff_quant

            is_onediff_quant_available = True
        except:
            pass

        torch_model = self.model.diffusion_model._deployable_module_model._torch_module
        for name, module in torch_model.named_modules():
            if isinstance(module, CrossAttention) and hasattr(module, "to_qkv"):
                # TODO(): support bias
                assert module.to_qkv.bias is None
                to_q_w_name = f"diffusion_model.{name}.to_q.weight"
                to_k_w_name = f"diffusion_model.{name}.to_k.weight"
                to_v_w_name = f"diffusion_model.{name}.to_v.weight"
                if (
                    to_q_w_name not in patches
                    or to_k_w_name not in patches
                    or to_v_w_name not in patches
                ):
                    continue
                to_q_w = patches[to_q_w_name][1]
                to_k_w = patches[to_k_w_name][1]
                to_v_w = patches[to_v_w_name][1]
                assert to_q_w[2] == to_k_w[2] and to_q_w[2] == to_v_w[2]
                to_qkv_w_name = f"diffusion_model.{name}.to_qkv.weight"

                dim_head = module.to_qkv.out_features // module.heads // 3
                tmp_list = [
                    torch.stack((to_q_w[0], to_k_w[0], to_v_w[0]), dim=0).reshape(
                        3, module.heads, dim_head, -1
                    ),  # (3, H, K, (BM))
                    torch.stack((to_q_w[1], to_k_w[1], to_v_w[1]), dim=0),
                ] + list(to_q_w[2:])

                patch_type = "onediff_int8"
                patch_value = tuple(tmp_list + [module])
                patches[to_qkv_w_name] = (patch_type, patch_value)

            if is_onediff_quant_available:
                if isinstance(
                    module, onediff_quant.DynamicQuantLinearModule
                ) or isinstance(module, onediff_quant.DynamicQuantConvModule):
                    w_name = f"diffusion_model.{name}.weight"
                    if w_name in patches:
                        patch_type = "onediff_int8"
                        patch_value = tuple(list(patches[w_name][1]) + [module])
                        patches[w_name] = (patch_type, patch_value)
                    b_name = f"diffusion_model.{name}.bias"
                    if b_name in patches:
                        patch_type = "onediff_int8"
                        patch_value = tuple(list(patches[b_name][1]) + [module])
                        patches[b_name] = (patch_type, patch_value)

        p = set()
        for k in patches:
            if k in self.model_keys:
                p.add(k)
                current_patches = self.patches.get(k, [])
                current_patches.append((strength_patch, patches[k], strength_model))
                self.patches[k] = current_patches

        return list(p)

    def calculate_weight(self, patches, weight, key):
        is_onediff_quant_available = False
        try:
            import onediff_quant

            is_onediff_quant_available = True
        except:
            pass
        for p in patches:
            alpha = p[0]
            v = p[1]
            strength_model = p[2]

            if strength_model != 1.0:
                weight *= strength_model

            if isinstance(v, list):
                v = (self.calculate_weight(v[1:], v[0].clone(), key),)

            if len(v) == 1:
                patch_type = "diff"
            elif len(v) == 2:
                patch_type = v[0]
                v = v[1]

            if patch_type == "diff":
                w1 = v[0]
                if alpha != 0.0:
                    if w1.shape != weight.shape:
                        print(
                            "WARNING SHAPE MISMATCH {} WEIGHT NOT MERGED {} != {}".format(
                                key, w1.shape, weight.shape
                            )
                        )
                    else:
                        weight += alpha * comfy.model_management.cast_to_device(
                            w1, weight.device, weight.dtype
                        )
            elif patch_type == "lora":  # lora/locon
                mat1 = comfy.model_management.cast_to_device(
                    v[0], weight.device, torch.float32
                )
                mat2 = comfy.model_management.cast_to_device(
                    v[1], weight.device, torch.float32
                )
                if v[2] is not None:
                    alpha *= v[2] / mat2.shape[0]
                if v[3] is not None:
                    # locon mid weights, hopefully the math is fine because I didn't properly test it
                    mat3 = comfy.model_management.cast_to_device(
                        v[3], weight.device, torch.float32
                    )
                    final_shape = [
                        mat2.shape[1],
                        mat2.shape[0],
                        mat3.shape[2],
                        mat3.shape[3],
                    ]
                    mat2 = (
                        torch.mm(
                            mat2.transpose(0, 1).flatten(start_dim=1),
                            mat3.transpose(0, 1).flatten(start_dim=1),
                        )
                        .reshape(final_shape)
                        .transpose(0, 1)
                    )
                try:
                    weight += (
                        (
                            alpha
                            * torch.mm(
                                mat1.flatten(start_dim=1), mat2.flatten(start_dim=1)
                            )
                        )
                        .reshape(weight.shape)
                        .type(weight.dtype)
                    )
                except Exception as e:
                    print("ERROR", key, e)
            elif patch_type == "lokr":
                w1 = v[0]
                w2 = v[1]
                w1_a = v[3]
                w1_b = v[4]
                w2_a = v[5]
                w2_b = v[6]
                t2 = v[7]
                dim = None

                if w1 is None:
                    dim = w1_b.shape[0]
                    w1 = torch.mm(
                        comfy.model_management.cast_to_device(
                            w1_a, weight.device, torch.float32
                        ),
                        comfy.model_management.cast_to_device(
                            w1_b, weight.device, torch.float32
                        ),
                    )
                else:
                    w1 = comfy.model_management.cast_to_device(
                        w1, weight.device, torch.float32
                    )

                if w2 is None:
                    dim = w2_b.shape[0]
                    if t2 is None:
                        w2 = torch.mm(
                            comfy.model_management.cast_to_device(
                                w2_a, weight.device, torch.float32
                            ),
                            comfy.model_management.cast_to_device(
                                w2_b, weight.device, torch.float32
                            ),
                        )
                    else:
                        w2 = torch.einsum(
                            "i j k l, j r, i p -> p r k l",
                            comfy.model_management.cast_to_device(
                                t2, weight.device, torch.float32
                            ),
                            comfy.model_management.cast_to_device(
                                w2_b, weight.device, torch.float32
                            ),
                            comfy.model_management.cast_to_device(
                                w2_a, weight.device, torch.float32
                            ),
                        )
                else:
                    w2 = comfy.model_management.cast_to_device(
                        w2, weight.device, torch.float32
                    )

                if len(w2.shape) == 4:
                    w1 = w1.unsqueeze(2).unsqueeze(2)
                if v[2] is not None and dim is not None:
                    alpha *= v[2] / dim

                try:
                    weight += alpha * torch.kron(w1, w2).reshape(weight.shape).type(
                        weight.dtype
                    )
                except Exception as e:
                    print("ERROR", key, e)
            elif patch_type == "loha":
                w1a = v[0]
                w1b = v[1]
                if v[2] is not None:
                    alpha *= v[2] / w1b.shape[0]
                w2a = v[3]
                w2b = v[4]
                if v[5] is not None:  # cp decomposition
                    t1 = v[5]
                    t2 = v[6]
                    m1 = torch.einsum(
                        "i j k l, j r, i p -> p r k l",
                        comfy.model_management.cast_to_device(
                            t1, weight.device, torch.float32
                        ),
                        comfy.model_management.cast_to_device(
                            w1b, weight.device, torch.float32
                        ),
                        comfy.model_management.cast_to_device(
                            w1a, weight.device, torch.float32
                        ),
                    )

                    m2 = torch.einsum(
                        "i j k l, j r, i p -> p r k l",
                        comfy.model_management.cast_to_device(
                            t2, weight.device, torch.float32
                        ),
                        comfy.model_management.cast_to_device(
                            w2b, weight.device, torch.float32
                        ),
                        comfy.model_management.cast_to_device(
                            w2a, weight.device, torch.float32
                        ),
                    )
                else:
                    m1 = torch.mm(
                        comfy.model_management.cast_to_device(
                            w1a, weight.device, torch.float32
                        ),
                        comfy.model_management.cast_to_device(
                            w1b, weight.device, torch.float32
                        ),
                    )
                    m2 = torch.mm(
                        comfy.model_management.cast_to_device(
                            w2a, weight.device, torch.float32
                        ),
                        comfy.model_management.cast_to_device(
                            w2b, weight.device, torch.float32
                        ),
                    )

                try:
                    weight += (alpha * m1 * m2).reshape(weight.shape).type(weight.dtype)
                except Exception as e:
                    print("ERROR", key, e)
            elif patch_type == "glora":
                if v[4] is not None:
                    alpha *= v[4] / v[0].shape[0]

                a1 = comfy.model_management.cast_to_device(
                    v[0].flatten(start_dim=1), weight.device, torch.float32
                )
                a2 = comfy.model_management.cast_to_device(
                    v[1].flatten(start_dim=1), weight.device, torch.float32
                )
                b1 = comfy.model_management.cast_to_device(
                    v[2].flatten(start_dim=1), weight.device, torch.float32
                )
                b2 = comfy.model_management.cast_to_device(
                    v[3].flatten(start_dim=1), weight.device, torch.float32
                )

                weight += (
                    (
                        (
                            torch.mm(b2, b1)
                            + torch.mm(torch.mm(weight.flatten(start_dim=1), a2), a1)
                        )
                        * alpha
                    )
                    .reshape(weight.shape)
                    .type(weight.dtype)
                )
            elif patch_type == "onediff_int8":
                # import pdb;pdb.set_trace()
                is_rewrite_qkv = True if "to_qkv" in key else False
                is_quant = False
                if (
                    is_onediff_quant_available
                    and len(v) == 5
                    and (
                        isinstance(v[4], onediff_quant.DynamicQuantLinearModule)
                        or isinstance(v[4], onediff_quant.DynamicQuantConvModule)
                    )
                ):
                    is_quant = True
                    org_weight_scale = (
                        v[4]
                        .weight_scale.reshape(
                            [-1] + [1 for _ in range(len(weight.shape) - 1)]
                        )
                        .to(weight.device)
                    )
                    weight = weight.to(torch.float32) * org_weight_scale

                mat1 = comfy.model_management.cast_to_device(
                    v[0], weight.device, torch.float32
                )
                mat2 = comfy.model_management.cast_to_device(
                    v[1], weight.device, torch.float32
                )
                if v[2] is not None:
                    if is_rewrite_qkv:
                        alpha *= v[2] / mat2.shape[1]
                    else:
                        alpha *= v[2] / mat2.shape[0]
                if v[3] is not None:
                    # TODO(): support rewrite qkv
                    assert not is_rewrite_qkv

                    # locon mid weights, hopefully the math is fine because I didn't properly test it
                    mat3 = comfy.model_management.cast_to_device(
                        v[3], weight.device, torch.float32
                    )
                    final_shape = [
                        mat2.shape[1],
                        mat2.shape[0],
                        mat3.shape[2],
                        mat3.shape[3],
                    ]
                    mat2 = (
                        torch.mm(
                            mat2.transpose(0, 1).flatten(start_dim=1),
                            mat3.transpose(0, 1).flatten(start_dim=1),
                        )
                        .reshape(final_shape)
                        .transpose(0, 1)
                    )
                try:
                    if is_rewrite_qkv:
                        heads = mat1.shape[1]
                        qkv_lora = alpha * torch.bmm(
                            mat1.reshape(3, -1, mat2.shape[1]),
                            mat2.flatten(start_dim=2),
                        )
                        qkv_lora = qkv_lora.reshape(
                            3, heads, -1, qkv_lora.shape[2]
                        )  # reshape to (3, H, K, (BM))
                        qkv_lora = qkv_lora.permute(
                            1, 0, 2, 3
                        )  # permute to (H, 3, K, (BM))
                        weight += qkv_lora.reshape(weight.shape).type(weight.dtype)
                    else:
                        weight += (
                            (
                                alpha
                                * torch.mm(
                                    mat1.flatten(start_dim=1), mat2.flatten(start_dim=1)
                                )
                            )
                            .reshape(weight.shape)
                            .type(weight.dtype)
                        )
                    if is_quant:
                        weight_max = torch.max(
                            weight.reshape(weight.shape[0], -1), dim=1
                        )[0].reshape([-1] + [1 for _ in range(len(weight.shape) - 1)])
                        weight_scale = torch.abs(weight_max) / 127
                        weight = torch.clamp(
                            torch.round(weight / weight_scale), -127, 127
                        ).to(weight.dtype)
                        weight_acc = (weight * weight_scale).sum(
                            dim=[d for d in range(1, len(weight.shape))]
                        )
                        weight_scale = weight_scale.reshape(v[4].weight_scale.shape).to(
                            v[4].weight_scale.dtype
                        )
                        weight_acc = weight_acc.reshape(v[4].weight_acc.shape).to(
                            v[4].weight_acc.dtype
                        )
                        v[4].weight_scale.copy_(weight_scale)
                        v[4].weight_acc.copy_(weight_acc)
                except Exception as e:
                    print("ERROR", key, e)
        return weight


class OneFlowDeepCacheSpeedUpModelPatcher(OneFlowSpeedUpModelPatcher):
    def __init__(
        self,
        model,
        load_device,
        offload_device,
        cache_layer_id,
        cache_block_id,
        size=0,
        current_device=None,
        weight_inplace_update=False,
        *,
        use_graph=None,
        gen_compile_options=None,
    ):
        from onediff.infer_compiler import (
            DeployableModule,
            oneflow_compile,
            OneflowCompileOptions,
        )

        self.weight_inplace_update = weight_inplace_update
        self.object_patches = {}
        self.object_patches_backup = {}
        self.size = size
        self.model = copy.copy(model)
        self.model.__dict__["_modules"] = copy.copy(model.__dict__["_modules"])
        self.deep_cache_unet = DeepCacheUNet(
            self.model.diffusion_model, cache_layer_id, cache_block_id
        )

        self.fast_deep_cache_unet = FastDeepCacheUNet(
            self.model.diffusion_model, cache_layer_id, cache_block_id
        )
        if use_graph:
            gen_compile_options = gen_compile_options or (
                lambda x: OneflowCompileOptions()
            )
            compile_options = gen_compile_options(self.deep_cache_unet)
            compile_options.use_graph = use_graph
            self.deep_cache_unet = oneflow_compile(
                self.deep_cache_unet,
                options=compile_options,
            )
            compile_options = gen_compile_options(self.fast_deep_cache_unet)
            compile_options.use_graph = use_graph
            self.fast_deep_cache_unet = oneflow_compile(
                self.fast_deep_cache_unet,
                options=compile_options,
            )
            self.model._register_state_dict_hook(state_dict_hook)

        self.patches = {}
        self.backup = {}
        self.model_options = {"transformer_options": {}}
        self.model_size()
        self.load_device = load_device
        self.offload_device = offload_device
        if current_device is None:
            self.current_device = self.offload_device
        else:
            self.current_device = current_device

        self.model_lowvram = getattr(model, "model_lowvram", False)


def get_mixed_speedup_class(module_cls):
    class MixedSpeedUpModelPatcher(OneFlowSpeedUpModelPatcher, module_cls):
        def __init__(
            self,
            model_patcher,
            load_device,
            offload_device,
            size=0,
            current_device=None,
            weight_inplace_update=False,
            *,
            use_graph=None,
        ):
            self.__dict__.update(**model_patcher.__dict__)
            OneFlowSpeedUpModelPatcher.__init__(
                self,
                model_patcher.model,
                load_device,
                offload_device,
                size,
                current_device,
                weight_inplace_update,
                use_graph=use_graph,
            )

        def clone(self):
            cloned = module_cls.clone(self)
            n = OneFlowSpeedUpModelPatcher(
                cloned.model,
                self.load_device,
                self.offload_device,
                self.size,
                self.current_device,
                weight_inplace_update=self.weight_inplace_update,
            )
            n.patches = {}
            for k in self.patches:
                n.patches[k] = self.patches[k][:]

            n.object_patches = self.object_patches.copy()
            n.model_options = copy.deepcopy(self.model_options)
            n.model_keys = self.model_keys

            n.model_lowvram = self.model_lowvram
            return n

    return MixedSpeedUpModelPatcher



from abc import abstractmethod

import comfy
import oneflow as th  # 'th' is the way ComfyUI name the torch
import oneflow.nn.functional as F
from einops import rearrange
from onediff.infer_compiler.backends.oneflow.transform import proxy_class, transform_mgr

onediff_comfy = transform_mgr.transform_package("comfy")

ops = onediff_comfy.ops.disable_weight_init
ResBlock = onediff_comfy.ldm.modules.diffusionmodules.openaimodel.ResBlock
checkpoint = onediff_comfy.ldm.modules.diffusionmodules.util.checkpoint


class Upsample(proxy_class(comfy.ldm.modules.diffusionmodules.openaimodel.Upsample)):
    # https://github.com/comfyanonymous/ComfyUI/blob/b0aab1e4ea3dfefe09c4f07de0e5237558097e22/comfy/ldm/modules/diffusionmodules/openaimodel.py#L82
    def forward(self, x, output_shape=None):
        assert x.shape[1] == self.channels
        if output_shape is not None and isinstance(output_shape, th.Tensor):
            if self.dims == 3:
                raise ValueError("output_shape shoud not be Tensor for dims == 3")
            else:
                x = F.interpolate_like(x, like=output_shape, mode="nearest")
        else:
            if self.dims == 3:
                shape = [x.shape[2], x.shape[3] * 2, x.shape[4] * 2]
                if output_shape is not None:
                    shape[1] = output_shape[3]
                    shape[2] = output_shape[4]
            else:
                shape = [x.shape[2] * 2, x.shape[3] * 2]
                if output_shape is not None:
                    shape[0] = output_shape[2]
                    shape[1] = output_shape[3]

            x = F.interpolate(x, size=shape, mode="nearest")

        if self.use_conv:
            x = self.conv(x)
        return x


class UNetModel(proxy_class(comfy.ldm.modules.diffusionmodules.openaimodel.UNetModel)):
    # https://github.com/comfyanonymous/ComfyUI/blob/b0aab1e4ea3dfefe09c4f07de0e5237558097e22/comfy/ldm/modules/diffusionmodules/openaimodel.py#L823
    def forward(
        self,
        x,
        timesteps=None,
        context=None,
        y=None,
        control=None,
        transformer_options={},
        **kwargs
    ):
        timestep_embedding = (
            onediff_comfy.ldm.modules.diffusionmodules.util.timestep_embedding
        )
        forward_timestep_embed = (
            onediff_comfy.ldm.modules.diffusionmodules.openaimodel.forward_timestep_embed
        )
        apply_control = (
            onediff_comfy.ldm.modules.diffusionmodules.openaimodel.apply_control
        )

        transformer_options["original_shape"] = list(x.shape)
        transformer_options["transformer_index"] = 0
        transformer_patches = transformer_options.get("patches", {})

        num_video_frames = kwargs.get("num_video_frames", self.default_num_video_frames)
        default_image_only_indicator = getattr(
            self, "default_image_only_indicator", None
        )
        # https://github.com/comfyanonymous/ComfyUI/blob/cb7c3a2921cfc0805be0229b4634e1143d60e6fe/comfy/ldm/modules/diffusionmodules/openaimodel.py#L832
        image_only_indicator = kwargs.get(
            "image_only_indicator", default_image_only_indicator
        )
        time_context = kwargs.get("time_context", None)

        assert (y is not None) == (
            self.num_classes is not None
        ), "must specify y if and only if the model is class-conditional"
        hs = []
        t_emb = timestep_embedding(
            timesteps, self.model_channels, repeat_only=False
        ).to(x.dtype)
        emb = self.time_embed(t_emb)

        if self.num_classes is not None:
            assert y.shape[0] == x.shape[0]
            emb = emb + self.label_emb(y)

        h = x
        for id, module in enumerate(self.input_blocks):
            transformer_options["block"] = ("input", id)
            h = forward_timestep_embed(
                module,
                h,
                emb,
                context,
                transformer_options,
                time_context=time_context,
                num_video_frames=num_video_frames,
                image_only_indicator=image_only_indicator,
            )
            h = apply_control(h, control, "input")
            if "input_block_patch" in transformer_patches:
                patch = transformer_patches["input_block_patch"]
                for p in patch:
                    h = p(h, transformer_options)

            hs.append(h)
            if "input_block_patch_after_skip" in transformer_patches:
                patch = transformer_patches["input_block_patch_after_skip"]
                for p in patch:
                    h = p(h, transformer_options)

        transformer_options["block"] = ("middle", 0)
        h = forward_timestep_embed(
            self.middle_block,
            h,
            emb,
            context,
            transformer_options,
            time_context=time_context,
            num_video_frames=num_video_frames,
            image_only_indicator=image_only_indicator,
        )
        h = apply_control(h, control, "middle")

        for id, module in enumerate(self.output_blocks):
            transformer_options["block"] = ("output", id)
            hsp = hs.pop()
            hsp = apply_control(hsp, control, "output")

            if "output_block_patch" in transformer_patches:
                patch = transformer_patches["output_block_patch"]
                for p in patch:
                    h, hsp = p(h, hsp, transformer_options)

            h = th.cat([h, hsp], dim=1)
            del hsp
            if len(hs) > 0:
                # output_shape = hs[-1].shape
                output_shape = hs[-1]
            else:
                output_shape = None
            h = forward_timestep_embed(
                module,
                h,
                emb,
                context,
                transformer_options,
                output_shape,
                time_context=time_context,
                num_video_frames=num_video_frames,
                image_only_indicator=image_only_indicator,
            )
        h = h.type(x.dtype)
        if self.predict_codebook_ids:
            return self.id_predictor(h)
        else:
            return self.out(h)


class VideoResBlock(
    proxy_class(comfy.ldm.modules.diffusionmodules.openaimodel.VideoResBlock)
):
    def _forward(self, x, emb):
        if self.updown:
            in_rest, in_conv = self.in_layers[:-1], self.in_layers[-1]
            h = in_rest(x)
            h = self.h_upd(h)
            x = self.x_upd(x)
            h = in_conv(h)
        else:
            h = self.in_layers(x)

        emb_out = None
        if not self.skip_t_emb:
            emb_out = self.emb_layers(emb).type(h.dtype)
            while len(emb_out.shape) < len(h.shape):
                emb_out = emb_out[..., None]
        if self.use_scale_shift_norm:
            out_norm, out_rest = self.out_layers[0], self.out_layers[1:]
            h = out_norm(h)
            if emb_out is not None:
                scale, shift = th.chunk(emb_out, 2, dim=1)
                h *= 1 + scale
                h += shift
            h = out_rest(h)
        else:
            if emb_out is not None:
                if self.exchange_temb_dims:
                    # emb_out = rearrange(emb_out, "b t c ... -> b c t ...")
                    # Rewrite for onediff SVD dynamic shape
                    emb_out = emb_out.permute(0, 2, 1, 3, 4)
                h = h + emb_out
            h = self.out_layers(h)
        return self.skip_connection(x) + h

    def forward(
        self,
        x: th.Tensor,
        emb: th.Tensor,
        num_video_frames: int,
        image_only_indicator=None,
    ) -> th.Tensor:
        # Rewrite for onediff SVD dynamic shape
        # x = super().forward(x, emb)
        x = checkpoint(self._forward, (x, emb), self.parameters(), self.use_checkpoint)

        # x_mix = rearrange(x, "(b t) c h w -> b c t h w", t=num_video_frames)
        batch_frames, _, _, _ = x.shape
        batch_size = batch_frames // num_video_frames
        x_mix = x.unflatten(0, shape=(batch_size, -1)).permute(0, 2, 1, 3, 4)
        # x = rearrange(x, "(b t) c h w -> b c t h w", t=num_video_frames)
        x = x.unflatten(0, shape=(batch_size, -1)).permute(0, 2, 1, 3, 4)

        # x = self.time_stack(
        #     x, rearrange(emb, "(b t) ... -> b t ...", t=num_video_frames)
        # )
        x = self.time_stack(x, emb.unflatten(0, shape=(batch_size, -1)))

        x = self.time_mixer(
            x_spatial=x_mix, x_temporal=x, image_only_indicator=image_only_indicator
        )
        # x = rearrange(x, "b c t h w -> (b t) c h w")
        x = x.permute(0, 2, 1, 3, 4).flatten(0, 1)
        return x



"""
Testing inference speed
save graph compiled example: python3 examples/unet_torch_interplay.py --save --model_id xx
load graph compiled example: python3 examples/unet_torch_interplay.py --load
"""
import importlib.metadata
import os
import random

import click

import torch
from packaging import version
import oneflow as flow  # usort: skip

from dataclasses import dataclass, fields

from onediff.infer_compiler import oneflow_compile
from tqdm import tqdm


@dataclass
class TensorInput(object):
    noise: torch.float16
    time: torch.int64
    cross_attention_dim: torch.float16

    @classmethod
    def gettype(cls, key):
        field_types = {field.name: field.type for field in fields(TensorInput)}
        return field_types[key]


def get_unet(token, _model_id, variant):
    from diffusers import UNet2DConditionModel

    unet = UNet2DConditionModel.from_pretrained(
        _model_id,
        use_auth_token=token,
        variant=variant,
        torch_dtype=torch.float16,
        subfolder="unet",
    )
    with torch.no_grad():
        unet = unet.to("cuda")
    return unet


def warmup_with_arg(graph, arg_meta_of_sizes, added):
    for arg_metas in arg_meta_of_sizes:
        print(f"warmup {arg_metas=}")
        arg_tensors = [
            torch.empty(arg_metas.noise, dtype=arg_metas.gettype("noise")).to("cuda"),
            torch.empty(arg_metas.time, dtype=arg_metas.gettype("time")).to("cuda"),
            torch.empty(
                arg_metas.cross_attention_dim,
                dtype=arg_metas.gettype("cross_attention_dim"),
            ).to("cuda"),
        ]
        graph(
            *arg_tensors, added_cond_kwargs=added, return_dict=False
        )  # build and warmup


def img_dim(i, start, stride):
    return start + stride * i


def noise_shape(batch_size, num_channels, image_w, image_h):
    sizes = (image_w // 8, image_h // 8)
    return (batch_size, num_channels) + sizes


def get_arg_meta_of_sizes(
    batch_sizes,
    resolution_scales,
    num_channels,
    cross_attention_dim,
    start=768,
    stride=128,
):
    return [
        TensorInput(
            noise_shape(
                batch_size,
                num_channels,
                img_dim(i, start, stride),
                img_dim(j, start, stride),
            ),
            (1,),
            (batch_size, 77, cross_attention_dim),
        )
        for batch_size in batch_sizes
        for i in resolution_scales
        for j in resolution_scales
    ]


@click.command()
@click.option("--token")
@click.option("--repeat", default=100)
@click.option("--sync_interval", default=50)
@click.option("--save", is_flag=True)
@click.option("--load", is_flag=True)
@click.option("--file", type=str, default="./unet_graphs")
@click.option(
    "--model_id", type=str, default="stabilityai/stable-diffusion-xl-base-1.0"
)
@click.option("--variant", type=str, default="fp16")
def benchmark(token, repeat, sync_interval, save, load, file, model_id, variant):
    RESOLUTION_SCALES = [2, 1, 0]
    BATCH_SIZES = [2]
    # TODO: reproduce bug caused by changing batch
    # BATCH_SIZES = [4, 2]

    unet = get_unet(token, model_id, variant)
    unet_graph = oneflow_compile(unet)

    num_channels = 4
    cross_attention_dim = unet.config["cross_attention_dim"]
    diffusers_version = version.parse(importlib.metadata.version("diffusers"))
    if diffusers_version < version.parse("0.21.0"):
        from diffusers.utils import floats_tensor
    else:
        from diffusers.utils.testing_utils import floats_tensor

    if (
        model_id == "stabilityai/stable-diffusion-xl-base-1.0"
        or "xl-base-1.0" in model_id
    ):
        # sdxl needed
        add_text_embeds = floats_tensor((2, 1280)).to("cuda").to(torch.float16)
        add_time_ids = floats_tensor((2, 6)).to("cuda").to(torch.float16)
        added_cond_kwargs = {"text_embeds": add_text_embeds, "time_ids": add_time_ids}
    else:
        added_cond_kwargs = None

    warmup_meta_of_sizes = get_arg_meta_of_sizes(
        batch_sizes=BATCH_SIZES,
        resolution_scales=RESOLUTION_SCALES,
        num_channels=num_channels,
        cross_attention_dim=cross_attention_dim,
    )
    for i, m in enumerate(warmup_meta_of_sizes):
        print(f"warmup case #{i + 1}:", m)

    # load graph from filepath
    if load:
        print("loading graphs...")
        unet_graph.load_graph(file)
    else:
        print("warmup with arguments...")
        warmup_with_arg(unet_graph, warmup_meta_of_sizes, added_cond_kwargs)

    # generate inputs with torch
    time_step = torch.tensor([10]).to("cuda")
    encoder_hidden_states_of_sizes = {
        batch_size: floats_tensor((batch_size, 77, cross_attention_dim))
        .to("cuda")
        .to(torch.float16)
        for batch_size in BATCH_SIZES
    }
    noise_of_sizes = [
        floats_tensor(arg_metas.noise).to("cuda").to(torch.float16)
        for arg_metas in warmup_meta_of_sizes
    ]
    flow._oneflow_internal.eager.Sync()
    import time

    # Testing inference speed
    t0 = time.time()
    for r in tqdm(range(repeat)):
        noise = random.choice(noise_of_sizes)
        encoder_hidden_states = encoder_hidden_states_of_sizes[noise.shape[0]]
        out = unet_graph(
            noise,
            time_step,
            encoder_hidden_states,
            added_cond_kwargs=added_cond_kwargs,
            return_dict=False,
        )
        if r == repeat - 1 or r % sync_interval == 0:
            flow._oneflow_internal.eager.Sync()
    assert isinstance(out[0], torch.Tensor)
    print(f"{type(out[0])=}")
    t1 = time.time()
    duration = t1 - t0
    throughput = repeat / duration
    print(
        f"Finish {repeat} steps in {duration:.3f} seconds, average {throughput:.2f}it/s"
    )

    # save graph to filepath
    if save:
        print("saving graphs...")
        unet_graph.save_graph(file)


if __name__ == "__main__":
    print(f"{flow.__path__=}")
    print(f"{flow.__version__=}")
    benchmark()



from .backend import OneDiffBackend
from .compile import get_compiled_graph
from .sd2 import SD21CompileCtx
from .utils import (
    get_onediff_backend,
    init_backend,
    is_nexfort_backend,
    is_oneflow_backend,
    OneDiffCompiledGraph,
)
from .vae import VaeCompileCtx

__all__ = [
    "get_compiled_graph",
    "SD21CompileCtx",
    "VaeCompileCtx",
    "OneDiffCompiledGraph",
    "OneDiffBackend",
    "get_onediff_backend",
    "is_oneflow_backend",
    "is_nexfort_backend",
    "init_backend",
]



from modules.sd_hijack import apply_optimizations

from onediff.infer_compiler import compile, oneflow_compile

from compile import OneDiffBackend

from .utils import (
    disable_unet_checkpointing,
    is_nexfort_backend,
    is_oneflow_backend,
    OneDiffCompiledGraph,
)


def get_compiled_graph(
    sd_model, unet_model=None, *, backend=None, quantization=None, options=None
) -> OneDiffCompiledGraph:
    diffusion_model = unet_model or sd_model.model.diffusion_model
    compiled_unet = onediff_compile(
        diffusion_model, backend=backend, quantization=quantization, options=options
    )
    return OneDiffCompiledGraph(sd_model, diffusion_model, compiled_unet, quantization)


def onediff_compile(
    unet_model,
    *,
    quantization: bool = False,
    backend: OneDiffBackend = None,
    options=None,
):
    if is_oneflow_backend(backend):
        return compile_unet_oneflow(
            unet_model, quantization=quantization, options=options
        )
    elif is_nexfort_backend(backend):
        return compile_unet_nexfort(
            unet_model, quantization=quantization, options=options
        )
    else:
        raise NotImplementedError(f"Can't find backend {backend} for OneDiff")


def compile_unet_oneflow(unet_model, *, quantization=False, options=None):
    from .oneflow.utils import init_oneflow_backend

    # 1. register mock map for converting torch to oneflow
    init_oneflow_backend()

    # 2. (for controlnet) remove attr forward to prevent mock failing
    if "forward" in unet_model.__dict__:
        unet_model.__dict__.pop("forward")

    # 3. disable checkpoint to prevent mock failing
    disable_unet_checkpointing(unet_model)

    compiled_unet_model = oneflow_compile(unet_model, options=options)
    if quantization:
        from .quantization import quant_unet_oneflow

        compiled_unet_model = quant_unet_oneflow(compiled_unet_model)
    return compiled_unet_model


def compile_unet_nexfort(unet_model, *, quantization=False, options=None):
    # TODO: support nexfort quant
    if quantization:
        raise NotImplementedError(
            "Quantization for nexfort backend is not implemented yet."
        )
    from .nexfort.utils import init_nexfort_backend

    init_nexfort_backend()
    apply_optimizations("nexfort")
    disable_unet_checkpointing(unet_model)
    unet_model.convert_to_fp16()
    return compile(unet_model, backend="nexfort", options=options)



"""
github: https://github.com/Kosinkadink/ComfyUI-AnimateDiff-Evolved
commit: 5d875d77fe6e31a4b0bc6dc36f0441eba3b6afe1
"""
import os

from onediff.infer_compiler.backends.oneflow.import_tools import DynamicModuleLoader
from onediff.infer_compiler.backends.oneflow.transform import transform_mgr

from ...sd_hijack_utils import Hijacker

COMFYUI_ROOT = os.getenv("COMFYUI_ROOT")
pkg_name = "ComfyUI-AnimateDiff-Evolved"
animatediff_root = os.path.join(COMFYUI_ROOT, "custom_nodes", pkg_name)
load_animatediff_package = True
try:
    if os.path.exists(animatediff_root):
        animatediff_pt = DynamicModuleLoader.from_path(animatediff_root)
        animatediff_of = transform_mgr.transform_package(pkg_name)
        comfy_of = transform_mgr.transform_package("comfy")
    else:
        load_animatediff_package = False
except Exception as e:
    print(f"Warning: Failed to load {pkg_name} from {animatediff_root} due to {e}")
    load_animatediff_package = False

animatediff_hijacker = Hijacker()



from collections import OrderedDict
from functools import singledispatch

import torch
from comfy.model_patcher import ModelPatcher
from comfy.sd import VAE
from onediff.torch_utils.module_operations import get_sub_module
from onediff.utils.import_utils import is_oneflow_available

from .._config import is_disable_oneflow_backend


@singledispatch
def switch_to_cached_model(new_model, cached_model):
    raise NotImplementedError(type(new_model))


@switch_to_cached_model.register
def _(new_model: ModelPatcher, cached_model):
    if type(new_model.model) != type(cached_model):
        raise TypeError(
            f"Model type mismatch: expected {type(cached_model)}, got {type(new_model.model)}"
        )

    cached_model.diffusion_model.load_state_dict(
        new_model.model.diffusion_model.state_dict(), strict=True
    )
    new_model.model.diffusion_model = cached_model.diffusion_model
    new_model.weight_inplace_update = True
    return new_model


@switch_to_cached_model.register
def _(new_model: VAE, cached_model):
    assert type(new_model.first_stage_model) == type(cached_model)
    for k, v in new_model.first_stage_model.state_dict().items():
        cached_v: torch.Tensor = get_sub_module(cached_model, k)
        assert v.dtype == cached_v.dtype
        cached_v.copy_(v)
    new_model.first_stage_model = cached_model
    return new_model


@singledispatch
def get_cached_model(model):
    return None
    # raise NotImplementedError(type(model))


@get_cached_model.register
def _(model: ModelPatcher):
    return model.model


@get_cached_model.register
def _(model: VAE):
    if is_oneflow_available() and not is_disable_oneflow_backend():
        from .oneflow.utils.booster_utils import is_using_oneflow_backend

        if is_using_oneflow_backend(model):
            return None

    # TODO(TEST) if support cache
    return model.first_stage_model


class BoosterCacheService:
    _cache = OrderedDict()

    def put(self, key, model):
        if key is None:
            return
        # oneflow backends output image error
        cached_model = get_cached_model(model)
        if cached_model:
            self._cache[key] = cached_model

    def get(self, key, default=None):
        return self._cache.get(key, default)

    def get_cached_model(self, key, model):
        cached_model = self.get(key, None)
        print(f"Cache lookup: Key='{key}', Cached Model Type='{type(cached_model)}'")
        if cached_model is not None:
            try:
                return switch_to_cached_model(model, cached_model)
            except Exception as e:
                print(f"An exception occurred when switching to cached model:")
                del self._cache[key]
                torch.cuda.empty_cache()

        return None



import argparse
import os
import sys
from datetime import datetime
from pathlib import Path

ONEDIFFBOX_ROOT = Path(os.path.abspath(__file__)).parents[0]
sys.path.insert(0, str(ONEDIFFBOX_ROOT))

from _logger import logger
from _utils import (
    build_image,
    calculate_sha256,
    gen_docker_compose_yaml,
    generate_docker_file,
    load_yaml,
    setup_repo,
)


def parse_args():
    parser = argparse.ArgumentParser(description="Build OneDiff Box")
    formatted_datetime = datetime.now().strftime("%Y%m%d-%H%M")

    parser.add_argument(
        "-y",
        "--yaml",
        type=str,
        default="config/community-default.yaml",
    )
    parser.add_argument(
        "-i",
        "--image",
        type=str,
        default="onediff",
    )
    parser.add_argument(
        "-t",
        "--tag",
        type=str,
        default=f"benchmark",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=".",
        help="the output directory of Dockerfile and Docker-compose file",
    )
    parser.add_argument(
        "-c",
        "--context",
        type=str,
        default=".",
        help="the path to build context",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="quiet mode",
    )
    args = parser.parse_args()
    return args


args = parse_args()


if __name__ == "__main__":
    image_config = load_yaml(file=args.yaml)
    file_hash = calculate_sha256(args.yaml)

    docker_file = generate_docker_file(
        args.yaml, file_hash, args.output, **image_config
    )
    version = os.path.splitext(os.path.basename(args.yaml))[0]
    image_name = f"{args.image}:{args.tag}-{version}"
    if not args.quiet:
        build_cmd = (
            f"docker build -f {docker_file} -t {args.image}:{args.tag} {args.context}"
        )
        print("Ready to build image by:")
        r = input("    " + build_cmd + " [y]/n ")
        if r == "" or r == "y" or r == "Y":
            logger.info(f"building image {image_name}")
            build_image(docker_file, image_name, args.context)
        else:
            print("building cancled")
    else:
        logger.info(f"building image {image_name}")
        build_image(docker_file, image_name, args.context)

    envs = image_config.pop("envs", [])
    volumes = image_config.pop(
        "volumes",
        [
            "$BENCHMARK_MODEL_PATH:/benchmark_model:ro",
        ],
    )
    compose_file, run_command = gen_docker_compose_yaml(
        f"onediff-benchmark-{version}", image_name, envs, volumes, args.output
    )
    logger.info(f"write docker-compose file to {compose_file}")
    logger.info(
        f"run container by:\n    {run_command}\n    and see {compose_file} for more"
    )



import logging
import os
from pathlib import Path

from onediff.utils.log_utils import ConfigurableLogger


def increment_path(path, exist_ok=False, sep="", mkdir=False):
    """
    Generates an incremented file or directory path if it exists, with optional mkdir; args: path, exist_ok=False,
    sep="", mkdir=False.

    Example: runs/exp --> runs/exp{sep}2, runs/exp{sep}3, ... etc
    """
    path = Path(path)  # os-agnostic
    if path.exists() and not exist_ok:
        path, suffix = (
            (path.with_suffix(""), path.suffix) if path.is_file() else (path, "")
        )

        # Method 1
        for n in range(2, 9999):
            p = f"{path}{sep}{n}{suffix}"  # increment path
            if not os.path.exists(p):  #
                break
        path = Path(p)

    if mkdir:
        path.mkdir(parents=True, exist_ok=True)  # make directory

    return path


logger = ConfigurableLogger()


def setup_logging(output_dir="results", exp_name="exp"):
    """
    Set up logging for an experiment.

    Args:
        output_dir (str): The base directory where experiment results will be stored.
        exp_name (str): The name of the experiment.

    Returns:
        tuple: A tuple containing the logger instance and the experiment directory path.
    """
    results_dir = Path(output_dir)
    exp_dir = increment_path(results_dir / exp_name, exist_ok=False, mkdir=True)
    # Configure logging
    logger.configure_logging(name="benchmarks", level=logging.INFO, log_dir=exp_dir)
    return logger, exp_dir



from comfy.model_patcher import ModelPatcher

from ..sd_hijack_utils import Hijacker
from .booster_utils import is_using_nexfort_backend
from .patch_management import create_patch_executor, PatchType


def clone_nexfort(org_fn, self, *args, **kwargs):
    n = org_fn(self, *args, **kwargs)
    create_patch_executor(PatchType.UiNodeWithIndexPatch).copy_to(self, n)
    dc_patch_executor = create_patch_executor(PatchType.DCUNetExecutorPatch)
    if dc_patch_executor.check_patch(self):
        dc_patch_executor.copy_to(self, n)
    return n


def cond_func(org_fn, self, *args, **kwargs):
    return is_using_nexfort_backend(self)


model_patch_hijacker = Hijacker()

model_patch_hijacker.register(ModelPatcher.clone, clone_nexfort, cond_func)



import hashlib
import os
import subprocess

import yaml

from _logger import logger

from git import Repo


def load_yaml(*, file):
    if not os.path.exists(file):
        raise RuntimeError(f"config file not existed: {file}")

    with open(file, "r") as file:
        yaml_content = yaml.safe_load(file)

    return yaml_content


def calculate_sha256(file_path):
    sha256_hash = hashlib.sha256()

    with open(file_path, "rb") as file:
        for chunk in iter(lambda: file.read(4096), b""):
            sha256_hash.update(chunk)
    return sha256_hash.hexdigest()


def setup_repo(repo_item: dict):
    if len(repo_item.keys()) != 1:
        raise RuntimeError(f"Only one key required, but got {repo_item.keys()}")
    else:
        for key in repo_item.keys():
            repo_name = key
            repo_item = repo_item.pop(repo_name)
            break

    repo_url = repo_item.pop("repo_url")
    branch = repo_item.pop("branch")
    commit = repo_item.pop("commit", None)
    cmds = repo_item.pop("cmds", None)

    repo_path = os.path.join(".", repo_name)
    if not os.path.exists(repo_path):
        logger.info(f"git clone {repo_url} to {repo_name}, branch: {branch}")
        git_repo = Repo.clone_from(repo_url, repo_path, branch=branch)
    else:
        logger.info(f"git repository {repo_name} has existed, use it")
        git_repo = Repo(repo_path)
    if commit is not None:
        git_repo.git.checkout(commit)
        logger.info(f"checkout {repo_name} to {commit}")
    docker_commands = f"COPY {repo_name} /app/{repo_name}"
    extra_cmds = ""
    if cmds is not None:
        extra_cmds = ["RUN ", " && \\\n".join(cmds)]
        extra_cmds = " ".join(extra_cmds)
        extra_cmds = "\n".join([f"WORKDIR /app/{repo_name}", extra_cmds])
    docker_commands = "\n".join([docker_commands, extra_cmds])

    return docker_commands


def generate_docker_file(yaml_file, file_hash, output_dir, **kwargs):
    image_config = kwargs

    base_image = image_config.pop("base_image", None)
    context_path = image_config.pop("context_path", None)
    oneflow_pip_index = image_config.pop("oneflow_pip_index", None)
    repos = image_config.pop("repos", None)
    proxy = image_config.pop("proxy", None)
    set_pip_mirror = image_config.pop("set_pip_mirror", None)

    origin_file_info = f"""#==== Generated from {yaml_file} ====
#     yaml file SHA256: {file_hash}
"""

    dockerfile_head = f"""
#==== Docker Base Image ====
FROM {base_image}
"""

    if set_pip_mirror is not None:
        dockerfile_set_pip_mirror = f"RUN {set_pip_mirror}"
    else:
        dockerfile_set_pip_mirror = ""

    dockerfile_oneflow = f"""
#==== Install the OneFlow ====
RUN pip install -f {oneflow_pip_index} oneflow
"""

    repos_cmds = []
    for repo in repos:
        repo = setup_repo(repo)
        repos_cmds.append(repo)
    repos_cmds = "\n\n".join(repos_cmds)
    dockerfile_repos = f"""
#==== Download and set up the repos ====
{repos_cmds}
"""

    docker_post_cmds = f"""
#==== Post setting
WORKDIR /app
"""

    dockerfile_content = "\n".join(
        [
            origin_file_info,
            dockerfile_head,
            dockerfile_set_pip_mirror,
            dockerfile_oneflow,
            dockerfile_repos,
            docker_post_cmds,
        ]
    )

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    dockerfile_name = os.path.join(output_dir, f"Dockerfile-{file_hash[0:8]}")
    logger.info(f"write Dockerfile to {dockerfile_name}")
    with open(dockerfile_name, "w") as f:
        f.write(dockerfile_content)
    return dockerfile_name


def build_image(docker_file, imagename, context):
    command = ["docker", "build", "-f", docker_file, "-t", imagename, context]
    try:
        process = subprocess.Popen(command, stdout=subprocess.PIPE, text=True)
        for line in iter(process.stdout.readline, ""):
            print(line, end="")  # Print each line
        process.wait()
    except subprocess.CalledProcessError as e:
        print(f"Command execution failed: {e}")


def gen_docker_compose_yaml(container_name, image, envs, volumes, output_dir):
    from collections import OrderedDict

    onediff_benchmark_service = {
        "container_name": None,
        "image": None,
        # "entrypoint": "sleep infinity",
        "tty": True,
        "stdin_open": True,
        "privileged": True,
        "shm_size": "8g",
        "network_mode": "host",
        "pids_limit": 2000,
        "cap_add": ["SYS_PTRACE"],
        "security_opt": ["seccomp=unconfined"],
        "environment": ["HF_HUB_OFFLINE=1"],
        "volumes": [],
        "working_dir": "/app",
        "restart": "no",
        "command": '/bin/bash -c "cd /app/onediff/benchmarks && bash run_benchmark.sh /benchmark_model"',
    }
    onediff_benchmark_service["container_name"] = container_name
    onediff_benchmark_service["image"] = image
    onediff_benchmark_service["environment"].extend(envs)
    onediff_benchmark_service["volumes"].extend(volumes)
    docker_compose_dict = {
        "version": "3.8",
        "services": {"onediff-benchmark": onediff_benchmark_service},
    }

    yaml_string = yaml.dump(docker_compose_dict)

    docker_compose_file = os.path.join(output_dir, f"docker-compose.{image}.yaml")

    docker_compose_readme = f"""#========
# run the OneDiff benchmark container by:
#    docker compose -f {docker_compose_file} up
#========

"""
    run_command = f"docker compose -f {docker_compose_file} up"
    with open(docker_compose_file, "w") as f:
        content = [docker_compose_readme, yaml_string]
        f.write("\n".join(content))
    return docker_compose_file, run_command



import os
from functools import wraps
from pathlib import Path
from textwrap import dedent
from zipfile import BadZipFile

import networks
import onediff_shared
from importlib_metadata import version

from onediff.utils.import_utils import is_oneflow_available

if is_oneflow_available():
    import oneflow as flow  # usort: skip

from compile import init_backend, is_oneflow_backend
from modules import shared
from modules.devices import torch_gc

from onediff.infer_compiler import DeployableModule

hints_message = dedent(
    """\
<div id="hintMessage" style="position: relative; padding: 20px; border: 1px solid #e0e0e0; border-radius: 5px; background-color: #f9f9f9;">
    <button onclick="document.getElementById('hintMessage').style.display = 'none'" style="position: absolute; top: 10px; right: 10px; background: none; border: none; font-size: 18px; cursor: pointer;">&times;</button>
    <div style="font-size: 18px; font-weight: bold; margin-bottom: 15px; color: #31708f;">
        Hints Message
    </div>
    <div style="padding: 10px; border: 1px solid #31708f; border-radius: 5px; background-color: #f9f9f9;">
        Hints: Enterprise function is not supported on your system.
    </div>
    <p style="margin-top: 15px;">
        If you need Enterprise Level Support for your system or business, please send an email to
        <a href="mailto:business@siliconflow.com" style="color: #31708f; text-decoration: none;">business@siliconflow.com</a>.
        <br>
        Tell us about your use case, deployment scale, and requirements.
    </p>
    <p>
        <strong>GitHub Issue:</strong>
        <a href="https://github.com/siliconflow/onediff/issues" style="color: #31708f; text-decoration: none;">https://github.com/siliconflow/onediff/issues</a>
    </p>
</div>
"""
)

all_compiler_caches = []


def all_compiler_caches_path():
    import modules.shared as shared

    # for webui <= 1.8, the function will be called before ui initialized
    # At that time, shared opts doesn't have attribute `onediff_compiler_caches_path`
    if not hasattr(shared.opts, "onediff_compiler_caches_path"):
        return None

    caches_path = Path(shared.opts.onediff_compiler_caches_path)
    if not caches_path.exists():
        caches_path.mkdir(parents=True)
    return shared.opts.onediff_compiler_caches_path


def get_all_compiler_caches():
    global all_compiler_caches
    if len(all_compiler_caches) == 0:
        refresh_all_compiler_caches()
    return all_compiler_caches


def refresh_all_compiler_caches(path: Path = None):
    global all_compiler_caches
    path = path or all_compiler_caches_path()
    if path is None:
        return

    all_compiler_caches = [f.stem for f in Path(path).iterdir() if f.is_file()]


def check_structure_change(current_type: dict[str, bool], model):
    return current_type != get_model_type(model)


def load_graph(compiled_unet: DeployableModule, compiler_cache: str):
    from compile import OneDiffCompiledGraph

    if isinstance(compiled_unet, OneDiffCompiledGraph):
        compiled_unet = compiled_unet.graph_module

    if compiler_cache in [None, "None"]:
        return

    compiler_cache_path = all_compiler_caches_path() + f"/{compiler_cache}"
    if not Path(compiler_cache_path).exists():
        raise FileNotFoundError(
            f"Cannot find cache {compiler_cache_path}, please make sure it exists"
        )
    try:
        compiled_unet.load_graph(compiler_cache_path, run_warmup=True)
    except BadZipFile:
        raise RuntimeError(
            "Load cache failed. Please make sure that the --disable-safe-unpickle parameter is added when starting the webui"
        )
    except Exception as e:
        raise RuntimeError(
            f"Load cache failed ({e}). Please make sure cache has the same sd version (or unet architure) with current checkpoint"
        )
    return compiled_unet


def save_graph(compiled_unet: DeployableModule, saved_cache_name: str = ""):
    from compile import OneDiffCompiledGraph

    if isinstance(compiled_unet, OneDiffCompiledGraph):
        compiled_unet = compiled_unet.graph_module

    if saved_cache_name in ["", None]:
        return

    if not os.access(str(all_compiler_caches_path()), os.W_OK):
        raise PermissionError(
            f"The directory {all_compiler_caches_path()} does not have write permissions, and compiler cache cannot be written to this directory. \
                                Please change it in the settings to a directory with write permissions"
        )
    if not Path(all_compiler_caches_path()).exists():
        Path(all_compiler_caches_path()).mkdir()
    saved_cache_name = all_compiler_caches_path() + f"/{saved_cache_name}"
    if not Path(saved_cache_name).exists():
        compiled_unet.save_graph(saved_cache_name)


def onediff_enabled_decorator(func):
    @wraps(func)
    def wrapper(
        self,
        p,
        quantization=False,
        compiler_cache=None,
        saved_cache_name="",
        always_recompile=False,
        backend=None,
    ):
        onediff_shared.onediff_enabled = True
        if networks.originals is not None:
            networks.originals.undo()
        init_backend(backend)
        try:
            return func(
                self,
                p,
                quantization=quantization,
                compiler_cache=compiler_cache,
                saved_cache_name=saved_cache_name,
                always_recompile=always_recompile,
                backend=backend,
            )
        finally:
            if networks.originals is not None:
                networks.originals.__init__()
            onediff_shared.onediff_enabled = False
            onediff_shared.previous_unet_type.update(**get_model_type(shared.sd_model))
            onediff_gc()

    return wrapper


def singleton_decorator(func):
    has_been_called = False

    def wrapper(*args, **kwargs):
        nonlocal has_been_called
        if not has_been_called:
            has_been_called = True
            return func(*args, **kwargs)

    return wrapper


def get_model_type(model):
    return {
        "is_sdxl": model.is_sdxl,
        "is_sd2": model.is_sd2,
        "is_sd1": model.is_sd1,
        "is_ssd": model.is_ssd,
    }


def onediff_gc():
    torch_gc()
    if is_oneflow_backend():
        flow.cuda.empty_cache()


def varify_can_use_quantization():
    try:
        import oneflow

        if version("oneflow") < "0.9.1":
            return False
    except ImportError as e:
        return False
    try:
        import onediff_quant
    except ImportError as e:
        return False
    return hasattr(oneflow._C, "dynamic_quantization")



from typing import List

import ldm.modules.attention
import onediff_shared
import sgm.modules.attention
import torch
from ldm.modules.diffusionmodules.util import timestep_embedding
from modules import shared
from modules.hypernetworks import hypernetwork
from modules.sd_hijack_optimizations import SdOptimization
from modules.sd_hijack_utils import CondFunc

from onediff.utils.import_utils import is_nexfort_available
from onediff_utils import singleton_decorator


@singleton_decorator
def init_nexfort_backend():
    CondFunc(
        "ldm.modules.diffusionmodules.openaimodel.UNetModel.forward",
        onediff_nexfort_unet_ldm_forward,
        lambda orig_func, *args, **kwargs: onediff_shared.onediff_enabled,
    )

    CondFunc(
        "sgm.modules.diffusionmodules.openaimodel.UNetModel.forward",
        onediff_nexfort_unet_sgm_forward,
        lambda orig_func, *args, **kwargs: onediff_shared.onediff_enabled,
    )

    def hijack_groupnorm32_forward(orig_func, self, x):
        return super(type(self), self).forward(x)
        # return self.forward(x)

    CondFunc(
        "ldm.modules.diffusionmodules.util.GroupNorm32.forward",
        hijack_groupnorm32_forward,
        lambda orig_func, *args, **kwargs: onediff_shared.onediff_enabled,
    )
    CondFunc(
        "sgm.modules.diffusionmodules.util.GroupNorm32.forward",
        hijack_groupnorm32_forward,
        lambda orig_func, *args, **kwargs: onediff_shared.onediff_enabled,
    )


@torch.autocast("cuda", enabled=False)
def onediff_nexfort_unet_sgm_forward(
    orig_func, self, x, timesteps=None, context=None, y=None, **kwargs
):
    assert (y is not None) == (
        self.num_classes is not None
    ), "must specify y if and only if the model is class-conditional"
    hs = []
    t_emb = timestep_embedding(timesteps, self.model_channels).half()
    emb = self.time_embed(t_emb)
    x = x.half()
    context = context.half() if context is not None else context
    y = y.half() if y is not None else y
    if self.num_classes is not None:
        assert y.shape[0] == x.shape[0]
        emb = emb + self.label_emb(y)
    h = x
    for module in self.input_blocks:
        h = module(h, emb, context)
        hs.append(h)
    h = self.middle_block(h, emb, context)
    for module in self.output_blocks:
        h = torch.cat([h, hs.pop()], dim=1)
        h = module(h, emb, context)
    h = h.type(x.dtype)
    return self.out(h)


@torch.autocast("cuda", enabled=False)
def onediff_nexfort_unet_ldm_forward(
    orig_func, self, x, timesteps=None, context=None, y=None, **kwargs
):
    assert (y is not None) == (
        self.num_classes is not None
    ), "must specify y if and only if the model is class-conditional"
    hs = []
    t_emb = timestep_embedding(timesteps, self.model_channels).half()
    emb = self.time_embed(t_emb)
    x = x.half()
    context = context.half()
    if self.num_classes is not None:
        assert y.shape[0] == x.shape[0]
        emb = emb + self.label_emb(y)
    h = x
    for module in self.input_blocks:
        h = module(h, emb, context)
        hs.append(h)
    h = self.middle_block(h, emb, context)
    for module in self.output_blocks:
        h = torch.cat([h, hs.pop()], dim=1)
        h = module(h, emb, context)
    if self.predict_codebook_ids:
        return self.id_predictor(h)
    else:
        return self.out(h)


def scaled_dot_product_attention_forward(self, x, context=None, mask=None, **kwargs):
    batch_size, sequence_length, inner_dim = x.shape

    if mask is not None:
        mask = self.prepare_attention_mask(mask, sequence_length, batch_size)
        mask = mask.view(batch_size, self.heads, -1, mask.shape[-1])

    h = self.heads
    q_in = self.to_q(x)

    context = x if context is None else context

    context_k, context_v = hypernetwork.apply_hypernetworks(
        shared.loaded_hypernetworks, context
    )
    k_in = self.to_k(context_k)
    v_in = self.to_v(context_v)

    head_dim = inner_dim // h
    q = q_in.view(batch_size, -1, h, head_dim).transpose(1, 2)
    k = k_in.view(batch_size, -1, h, head_dim).transpose(1, 2)
    v = v_in.view(batch_size, -1, h, head_dim).transpose(1, 2)

    del q_in, k_in, v_in

    dtype = q.dtype

    # the output of sdp = (batch, num_heads, seq_len, head_dim)
    hidden_states = torch.nn.functional.scaled_dot_product_attention(
        q, k, v, attn_mask=mask, dropout_p=0.0, is_causal=False
    )

    hidden_states = hidden_states.transpose(1, 2).reshape(batch_size, -1, h * head_dim)
    hidden_states = hidden_states.to(dtype)

    # linear proj
    hidden_states = self.to_out[0](hidden_states)
    # dropout
    hidden_states = self.to_out[1](hidden_states)
    return hidden_states


class SdOptimizationNexfort(SdOptimization):
    name = "nexfort"
    cmd_opt = "nexfort"
    priority = 10

    def is_available(self):
        return is_nexfort_available()

    def apply(self):
        ldm.modules.attention.CrossAttention.forward = (
            scaled_dot_product_attention_forward
        )
        sgm.modules.attention.CrossAttention.forward = (
            scaled_dot_product_attention_forward
        )


def add_nexfort_optimizer(res: List):
    res.append(SdOptimizationNexfort())



from typing import Any, Optional

import comfy
import oneflow as torch
import oneflow.nn as nn
import oneflow.nn.functional as F
from onediff.infer_compiler.backends.oneflow.transform import proxy_class


def Normalize(in_channels, num_groups=32):
    return torch.nn.GroupNorm(
        num_groups=num_groups, num_channels=in_channels, eps=1e-6, affine=True
    )



import argparse
import json
import os
import sys
import time
from io import BytesIO
from typing import List, NamedTuple, Union

import numpy as np
from PIL import Image
from skimage.metrics import structural_similarity as ssim

sys.path.append("./src")

from core.log_utils import setup_logging
from core.service_client import comfy_client_context
from input_registration import dispatch_generator

DEFAULT_HOST = "127.0.0.1"
DEFAULT_COMFY_PORT = "8188"


def parse_args():
    parser = argparse.ArgumentParser(description="Test ComfyUI workflow")
    parser.add_argument(
        "-w", "--workflow", type=str, nargs="+", required=True, help="Workflow file(s)"
    )
    parser.add_argument(
        "--listen", type=str, default=DEFAULT_HOST, help="Service listen address"
    )
    parser.add_argument(
        "--comfy-port",
        type=str,
        default=DEFAULT_COMFY_PORT,
        help="ComfyUI service port",
    )
    parser.add_argument(
        "--output-images", action="store_true", help="Enable output of images."
    )
    parser.add_argument(
        "--baseline-dir", type=str, help="Directory for baseline output."
    )
    parser.add_argument(
        "--ssim-threshold",
        type=float,
        default=0.5,
        help="SSIM threshold for image comparison.",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        default="results",
        help="Directory for output results.",
    )
    parser.add_argument(
        "--exp-name", type=str, default="exp", help="Experiment name for logging."
    )
    return parser.parse_args()


def save_image(image_data: bytes, output_dir: str, image_name: str) -> str:
    os.makedirs(output_dir, exist_ok=True)
    image_path = os.path.join(output_dir, image_name)
    Image.open(BytesIO(image_data)).save(image_path)
    return image_path


def calculate_ssim(image1: Image.Image, image2: Image.Image) -> float:
    image1_np = np.array(image1.convert("RGB"))
    image2_np = np.array(image2.convert("RGB"))
    assert image1_np.shape == image2_np.shape, "Images must have the same dimensions"
    return ssim(image1_np, image2_np, channel_axis=2)


class ImageInfo(NamedTuple):
    height: int
    width: int
    image_path: str = ""


class WorkflowProcessor:
    def __init__(
        self,
        output_images: bool,
        output_dir: str,
        logger,
    ):
        self.output_images = output_images
        self.output_dir = output_dir
        self.logger = logger

    def process_image(self, image_data: bytes, index: int) -> ImageInfo:
        pil_image = Image.open(BytesIO(image_data))
        self.logger.info(
            f"Image Size - Height: {pil_image.height}px, Width: {pil_image.width}px"
        )
        assert np.array(pil_image).any() != 0, "Image is blank"

        image_path = ""
        if self.output_images:
            image_path = save_image(image_data, self.output_dir, f"image_{index}.png")
            self.logger.info(f"Saved image to: {image_path}")

        return ImageInfo(
            height=pil_image.height, width=pil_image.width, image_path=image_path
        )


def run_workflow(
    workflow: Union[str, List[str]],
    comfy_port: str = DEFAULT_COMFY_PORT,
    output_images: bool = True,
    baseline_dir: str = None,
    ssim_threshold: float = 0.5,
    output_dir: str = "results",
    exp_name: str = "exp",
) -> None:
    logger, result_dir = setup_logging(output_dir, exp_name=exp_name)
    logger.info(f"Result directory: {result_dir}")

    processor = WorkflowProcessor(
        output_images,
        os.path.join(result_dir, "imgs"),
        logger,
    )

    result = {}
    result_file_name = "results.json"
    baseline_result = None
    if baseline_dir:
        result_file_path = os.path.join(baseline_dir, result_file_name)
        if os.path.exists(result_file_path):
            with open(result_file_path, "r") as fp:
                baseline_result = json.load(fp)

    with comfy_client_context(port=comfy_port) as client:
        logger.info(f"Testing workflows: {workflow}")

        for i, comfy_graph in enumerate(dispatch_generator(workflow)):
            start_time = time.time()
            images = client.get_images(comfy_graph.graph)
            end_time = time.time()

            if not images:
                logger.error("No images generated")
                raise ValueError("No images generated")
            if len(images) != 1:
                logger.error(
                    f"Expected 1 image, but got {len(images)} images. Batch Size == 1"
                )
                raise ValueError(
                    f"Expected 1 image, but got {len(images)} images. Batch Size == 1"
                )

            for images_output in images.values():
                for image_data in images_output:
                    out = processor.process_image(image_data, i)
                    result[i] = {
                        "height": out.height,
                        "width": out.width,
                        "image_path": out.image_path,
                    }
                    e2e_time = end_time - start_time
                    logger.info(f"Workflow {i} E2E:  {e2e_time:.2f} seconds")
                    result[i].update({"e2e_time": e2e_time})

                    if baseline_dir:
                        baseline_image_path = os.path.join(
                            baseline_dir, f"image_{i}.png"
                        )
                        baseline_image = Image.open(baseline_image_path)
                        pil_image = Image.open(BytesIO(image_data))
                        ssim_value = calculate_ssim(pil_image, baseline_image)
                        result[i].update(
                            {
                                "ssim_value": ssim_value,
                                "basic_image_path": baseline_image_path,
                            }
                        )
                        logger.info(f"SSIM: {ssim_value=}")
                        assert (
                            ssim_value > ssim_threshold
                        ), f"SSIM value {ssim_value} is not greater than the threshold {ssim_threshold}"

                        if baseline_result:
                            basic_time = baseline_result[str(i)]["e2e_time"]
                            result[i].update({"basic_e2e_time": basic_time})
                            percentage_improvement = (
                                basic_time - e2e_time
                            ) / basic_time
                            result[i].update(
                                {"percentage_improvement": percentage_improvement}
                            )

        # Save results to a JSON file
        result_file = os.path.join(result_dir, result_file_name)
        with open(result_file, "w") as f:
            json.dump(result, f, indent=4)

        logger.info(f"Results saved to {result_file}")


if __name__ == "__main__":
    args = parse_args()
    run_workflow(
        args.workflow,
        args.comfy_port,
        args.output_images,
        args.baseline_dir,
        args.ssim_threshold,
        args.output_dir,
        args.exp_name,
    )



import torch
from comfy import model_management

from ..booster_utils import clear_deployable_module_cache_and_unbind
from ..patch_management import create_patch_executor, PatchType

from .CrossAttentionPatch import Attn2Replace, ipadapter_attention


def set_model_patch_replace(
    org_fn, model, patch_kwargs, key, attention_func=ipadapter_attention
):
    diff_model = model.model.diffusion_model
    cache_patch_executor = create_patch_executor(PatchType.CachedCrossAttentionPatch)
    unet_extra_options_patch_executor = create_patch_executor(
        PatchType.UNetExtraInputOptions
    )
    cache_dict = cache_patch_executor.get_patch(diff_model)
    ui_cache_key = create_patch_executor(PatchType.UiNodeWithIndexPatch).get_patch(
        model
    )
    unet_extra_options = unet_extra_options_patch_executor.get_patch(diff_model)

    if "attn2" not in unet_extra_options:
        unet_extra_options["attn2"] = {}

    to = model.model_options["transformer_options"].copy()
    if "patches_replace" not in to:
        to["patches_replace"] = {}
    else:
        to["patches_replace"] = to["patches_replace"].copy()

    if "attn2" not in to["patches_replace"]:
        to["patches_replace"]["attn2"] = {}
    else:
        to["patches_replace"]["attn2"] = to["patches_replace"]["attn2"].copy()

    def split_patch_kwargs(patch_kwargs):
        split1dict = {}
        split2dict = {}
        for k, v in patch_kwargs.items():
            if k in ["cond", "cond_alt", "uncond", "mask", "weight"] or isinstance(
                v, torch.Tensor
            ):
                split1dict[k] = v
            else:
                split2dict[k] = v

        # patch for weight
        # weight = split1dict["weight"]
        # if isinstance(weight, (int, float)):
        #     weight = torch.tensor(weight)
        #     split1dict["weight"] = weight.to(model_management.get_torch_device())

        return split1dict, split2dict

    new_patch_kwargs, patch_kwargs = split_patch_kwargs(patch_kwargs)
    # update patch_kwargs
    if key in cache_dict:
        try:
            attn2_m = cache_dict[key]
            index = attn2_m.cache_map.get(ui_cache_key, None)
            if index is not None:
                unet_extra_options["attn2"][attn2_m.forward_patch_key][
                    index
                ] = new_patch_kwargs

                to["patches_replace"]["attn2"][key] = attn2_m
                model.model_options["transformer_options"] = to
                return

        except Exception as e:
            clear_deployable_module_cache_and_unbind(model)

    if key not in to["patches_replace"]["attn2"]:
        if key not in cache_dict:
            attn2_m = Attn2Replace(attention_func, **patch_kwargs)
            cache_dict[key] = attn2_m
            index = len(attn2_m.callback) - 1
            attn2_m.cache_map[ui_cache_key] = index
            unet_extra_options["attn2"][attn2_m.forward_patch_key] = [new_patch_kwargs]
        else:
            attn2_m = cache_dict[key]

        to["patches_replace"]["attn2"][key] = attn2_m
        model.model_options["transformer_options"] = to
    else:
        attn2_m: Attn2Replace = to["patches_replace"]["attn2"][key]
        unet_extra_options["attn2"][attn2_m.forward_patch_key].append(
            new_patch_kwargs
        )  # update last patch
        attn2_m.cache_map[ui_cache_key] = len(attn2_m.callback) - 1



# ComfyUI/custom_nodes/ComfyUI-AnimateDiff-Evolved/animatediff/motion_module_ad.py
import oneflow as torch
from einops import repeat
from onediff.infer_compiler.backends.oneflow.transform import register

from ._config import animatediff_of, animatediff_pt

TemporalTransformer3DModel_OF_CLS = (
    animatediff_of.animatediff.motion_module_ad.TemporalTransformer3DModel
)
TemporalTransformer3DModel_PT_CLS = (
    animatediff_pt.animatediff.motion_module_ad.TemporalTransformer3DModel
)
VersatileAttention_OF_CLS = (
    animatediff_of.animatediff.motion_module_ad.VersatileAttention
)
VersatileAttention_PT_CLS = (
    animatediff_pt.animatediff.motion_module_ad.VersatileAttention
)


class TemporalTransformer3DModel_OF(TemporalTransformer3DModel_OF_CLS):
    def get_cameractrl_effect(self, hidden_states: torch.Tensor):
        # if no raw camera_Ctrl, return None
        if self.raw_cameractrl_effect is None:
            return 1.0
        # if raw_cameractrl is not a Tensor, return it (should be a float)
        if type(self.raw_cameractrl_effect) != torch.Tensor:
            return self.raw_cameractrl_effect
        shape = hidden_states.shape
        batch, channel, height, width = shape
        # if temp_cameractrl already calculated, return it
        if self.temp_cameractrl_effect != None:
            # check if hidden_states batch matches
            if batch == self.prev_cameractrl_hidden_states_batch:
                if self.sub_idxs is not None:
                    return self.temp_cameractrl_effect[:, self.sub_idxs, :]
                return self.temp_cameractrl_effect
            # if does not match, reset cached temp_cameractrl and recalculate it
            del self.temp_cameractrl_effect
            self.temp_cameractrl_effect = None
        # otherwise, calculate temp_cameractrl
        self.prev_cameractrl_hidden_states_batch = batch
        mask = prepare_mask_batch(
            self.raw_scale_mask, shape=(self.full_length, 1, height, width)
        )
        mask = repeat_to_batch_size(mask, self.full_length)
        # if mask not the same amount length as full length, make it match
        if self.full_length != mask.shape[0]:
            mask = broadcast_image_to(mask, self.full_length, 1)
        # reshape mask to attention K shape (h*w, latent_count, 1)
        batch, channel, height, width = mask.shape
        # first, perform same operations as on hidden_states,
        # turning (b, c, h, w) -> (b, h*w, c)
        mask = mask.permute(0, 2, 3, 1).reshape(batch, height * width, channel)
        # then, make it the same shape as attention's k, (h*w, b, c)
        mask = mask.permute(1, 0, 2)
        # make masks match the expected length of h*w
        batched_number = shape[0] // self.video_length
        if batched_number > 1:
            mask = torch.cat([mask] * batched_number, dim=0)
        # cache mask and set to proper device
        self.temp_cameractrl_effect = mask
        # move temp_cameractrl to proper dtype + device
        self.temp_cameractrl_effect = self.temp_cameractrl_effect.to(
            dtype=hidden_states.dtype, device=hidden_states.device
        )
        # return subset of masks, if needed
        if self.sub_idxs is not None:
            return self.temp_cameractrl_effect[:, self.sub_idxs, :]
        return self.temp_cameractrl_effect

    def forward(
        self,
        hidden_states,
        encoder_hidden_states=None,
        attention_mask=None,
        view_options=None,
        mm_kwargs: dict[str] = None,
    ):
        batch, channel, height, width = hidden_states.shape
        residual = hidden_states
        cameractrl_effect = self.get_cameractrl_effect(hidden_states)

        scale_mask = self.get_scale_mask(hidden_states)
        # add some casts for fp8 purposes - does not affect speed otherwise
        hidden_states = self.norm(hidden_states).to(hidden_states.dtype)
        inner_dim = hidden_states.shape[1]
        # hidden_states = hidden_states.permute(0, 2, 3, 1).reshape(
        #     batch, height * width, inner_dim
        # )
        hidden_states = hidden_states.permute(0, 2, 3, 1).reshape(batch, -1, inner_dim)
        hidden_states = self.proj_in(hidden_states).to(hidden_states.dtype)

        # Transformer Blocks
        for block in self.transformer_blocks:
            hidden_states = block(
                hidden_states,
                encoder_hidden_states=encoder_hidden_states,
                attention_mask=attention_mask,
                video_length=self.video_length,
                scale_mask=scale_mask,
                cameractrl_effect=cameractrl_effect,
                view_options=view_options,
                mm_kwargs=mm_kwargs,
            )

        # output
        hidden_states = self.proj_out(hidden_states)
        # hidden_states = (
        #     hidden_states.reshape(batch, height, width, inner_dim)
        #     .permute(0, 3, 1, 2)
        #     .contiguous()
        # )
        # b (h w) i -> b i h w
        hidden_states = hidden_states.permute(0, 2, 1).reshape_as(residual)

        output = hidden_states + residual

        return output


class VersatileAttention_OF(VersatileAttention_OF_CLS):
    def forward(
        self,
        hidden_states,
        encoder_hidden_states=None,
        attention_mask=None,
        video_length=None,
        scale_mask=None,
        cameractrl_effect=1.0,
        mm_kwargs: dict[str] = {},
    ):
        if self.attention_mode != "Temporal":
            raise NotImplementedError
        d = hidden_states.shape[1]
        # hidden_states = rearrange(
        #     hidden_states, "(b f) d c -> (b d) f c", f=video_length
        # )
        # (b f) d c -> b f d c -> b d f c -> (b d) f c
        # print(f'in forward: hidden_states.shape: {hidden_states.shape}')
        b = hidden_states.shape[0] // video_length
        hidden_states = (
            hidden_states.unflatten(0, (b, -1)).permute(0, 2, 1, 3).flatten(0, 1)
        )

        if self.pos_encoder is not None:
            hidden_states = self.pos_encoder(hidden_states).to(hidden_states.dtype)

        encoder_hidden_states = (
            repeat(encoder_hidden_states, "b n c -> (b d) n c", d=d)
            if encoder_hidden_states is not None
            else encoder_hidden_states
        )
        if (
            self.camera_feature_enabled
            and self.qkv_merge is not None
            and mm_kwargs is not None
            and "camera_feature" in mm_kwargs
        ):
            camera_feature: torch.Tensor = mm_kwargs["camera_feature"]
            hidden_states = (
                self.qkv_merge(hidden_states + camera_feature) + hidden_states
            ) * cameractrl_effect + hidden_states * (1.0 - cameractrl_effect)

        # hidden_states = super().forward(
        #     hidden_states,
        #     encoder_hidden_states,
        #     value=None,
        #     mask=attention_mask,
        #     scale_mask=scale_mask,
        # )
        from .utils_motion import CrossAttentionMM_OF

        hidden_states = CrossAttentionMM_OF.forward(
            self,
            hidden_states,
            encoder_hidden_states,
            value=None,
            mask=attention_mask,
            scale_mask=scale_mask,
        )

        # hidden_states = rearrange(hidden_states, "(b d) f c -> (b f) d c", d=d)
        # (b d) f c -> b d f c -> b f d c -> (b f) d c
        hidden_states = (
            hidden_states.unflatten(0, (b, -1)).permute(0, 2, 1, 3).flatten(0, 1)
        )

        return hidden_states


register(
    torch2oflow_class_map={
        TemporalTransformer3DModel_PT_CLS: TemporalTransformer3DModel_OF
    }
)

# import torch as torch_pt
# from onediff.infer_compiler.backends.oneflow.transform import torch2oflow

# @torch2oflow.register(TemporalTransformer3DModel_PT_CLS)
# def _(mod, verbose=False):
#     of_mod = torch2oflow.dispatch(torch_pt.nn.Module)(mod, verbose)
#     of_mod.video_length = torch.tensor(mod.video_length)
#     return of_mod

register(torch2oflow_class_map={VersatileAttention_PT_CLS: VersatileAttention_OF})
